#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import osquery as rdf_osquery
from grr_response_proto import osquery_pb2


def ToProtoOsqueryArgs(rdf: rdf_osquery.OsqueryArgs) -> osquery_pb2.OsqueryArgs:
  return rdf.AsPrimitiveProto()


def ToRDFOsqueryArgs(proto: osquery_pb2.OsqueryArgs) -> rdf_osquery.OsqueryArgs:
  return rdf_osquery.OsqueryArgs.FromSerializedBytes(proto.SerializeToString())


def ToProtoOsqueryColumn(
    rdf: rdf_osquery.OsqueryColumn,
) -> osquery_pb2.OsqueryColumn:
  return rdf.AsPrimitiveProto()


def ToRDFOsqueryColumn(
    proto: osquery_pb2.OsqueryColumn,
) -> rdf_osquery.OsqueryColumn:
  return rdf_osquery.OsqueryColumn.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoOsqueryHeader(
    rdf: rdf_osquery.OsqueryHeader,
) -> osquery_pb2.OsqueryHeader:
  return rdf.AsPrimitiveProto()


def ToRDFOsqueryHeader(
    proto: osquery_pb2.OsqueryHeader,
) -> rdf_osquery.OsqueryHeader:
  return rdf_osquery.OsqueryHeader.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoOsqueryRow(rdf: rdf_osquery.OsqueryRow) -> osquery_pb2.OsqueryRow:
  return rdf.AsPrimitiveProto()


def ToRDFOsqueryRow(proto: osquery_pb2.OsqueryRow) -> rdf_osquery.OsqueryRow:
  return rdf_osquery.OsqueryRow.FromSerializedBytes(proto.SerializeToString())


def ToProtoOsqueryTable(
    rdf: rdf_osquery.OsqueryTable,
) -> osquery_pb2.OsqueryTable:
  return rdf.AsPrimitiveProto()


def ToRDFOsqueryTable(
    proto: osquery_pb2.OsqueryTable,
) -> rdf_osquery.OsqueryTable:
  return rdf_osquery.OsqueryTable.FromSerializedBytes(proto.SerializeToString())


def ToProtoOsqueryResult(
    rdf: rdf_osquery.OsqueryResult,
) -> osquery_pb2.OsqueryResult:
  return rdf.AsPrimitiveProto()


def ToRDFOsqueryResult(
    proto: osquery_pb2.OsqueryResult,
) -> rdf_osquery.OsqueryResult:
  return rdf_osquery.OsqueryResult.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoOsqueryProgress(
    rdf: rdf_osquery.OsqueryProgress,
) -> osquery_pb2.OsqueryProgress:
  return rdf.AsPrimitiveProto()


def ToRDFOsqueryProgress(
    proto: osquery_pb2.OsqueryProgress,
) -> rdf_osquery.OsqueryProgress:
  return rdf_osquery.OsqueryProgress.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""A module with utilities for retrying function execution."""

from collections.abc import Callable
import dataclasses
import datetime
import functools
import logging
import random
import time
from typing import Generic, Optional, TypeVar, Union


@dataclasses.dataclass
class Opts:
  """Options that customize the retry mechanism.

  Attributes:
    attempts: The number of attempts to retry the call.
    init_delay: An initial value for delay between retries.
    max_delay: A maximum value for delay between retries.
    backoff: A backoff multiplayer for the delay between retries.
    jitter: A random jitter to add to delay between retries.
    sleep: A sleep function used for delaying retries.
  """

  attempts: int = 1

  init_delay: datetime.timedelta = datetime.timedelta(0)
  max_delay: datetime.timedelta = datetime.timedelta.max
  backoff: float = 1.0
  jitter: float = 0.0

  # fmt: off
  sleep: Callable[[datetime.timedelta], None] = (
      lambda timedelta: time.sleep(timedelta.total_seconds())
  )
  # fmt: on


_E = TypeVar("_E", bound=Exception)


class On(Generic[_E]):
  """A decorator that retries the wrapped function on exception."""

  def __init__(
      self,
      exception: Union[type[_E], tuple[type[_E], ...]],
      opts: Optional[Opts] = None,
  ) -> None:
    """Initializes the decorator.

    Args:
      exception: A sequence of exceptions to retry on.
      opts: Options that customize the retry behaviour.
    """
    self._when = When(exception, lambda _: True, opts=opts)

  _R = TypeVar("_R")

  def __call__(self, func: Callable[..., _R]) -> Callable[..., _R]:
    """Wraps the specified function into a retryable function.

    The wrapped function will be attempted to be called specified number of
    times after which the error will be propagated.

    Args:
      func: A function to wrap.

    Returns:
      A wrapped function that retries on failures.
    """
    return self._when(func)


class When(Generic[_E]):
  """A decorator that retries function on exception if predicate is met."""

  def __init__(
      self,
      exception: Union[type[_E], tuple[type[_E], ...]],
      predicate: Callable[[_E], bool],
      opts: Optional[Opts] = None,
  ) -> None:
    """Initializes the decorator.

    Args:
      exception: An exception type to catch.
      predicate: A predicate to check whether to retry the exception.
      opts: Options that customize the retry behaviour.
    """
    if opts is None:
      opts = Opts()

    if opts.attempts < 1:
      raise ValueError("Non-positive number of retries")

    self._exception = exception
    self._predicate = predicate
    self._opts = opts

  _R = TypeVar("_R")

  def __call__(self, func: Callable[..., _R]) -> Callable[..., _R]:
    """Wraps the specified function into a retryable function.

    The wrapped function will be attempted to be called specified number of
    times after which the error will be propagated.

    Args:
      func: A function to wrap.

    Returns:
      A wrapped function that retries on failures.
    """
    opts = self._opts

    @functools.wraps(func)
    def Wrapped(*args, **kwargs) -> On._R:
      attempts = 0
      delay = opts.init_delay

      while True:
        try:
          return func(*args, **kwargs)
        except self._exception as error:
          attempts += 1
          if attempts == opts.attempts:
            raise

          if not self._predicate(error):
            raise

          jitter = random.uniform(-opts.jitter, +opts.jitter)
          jittered_delay = delay + delay * jitter

          logging.warning("'%s', to be retried in %s", error, jittered_delay)
          opts.sleep(jittered_delay)

          # Note that we calculate the new delay value basing on the base delay,
          # not the jittered one, otherwise the delay values might become too
          # unpredictable.
          delay = min(delay * opts.backoff, opts.max_delay)

    return Wrapped



#!/usr/bin/env python
"""Test for the flow state class."""

from absl import app

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import test_base as rdf_test_base
from grr.test_lib import test_lib


class SessionIDTest(rdf_test_base.RDFValueTestMixin, test_lib.GRRBaseTest):
  """Test SessionID."""

  rdfvalue_class = rdfvalue.SessionID

  def GenerateSample(self, number=0):
    id_str = "%08X" % (number % 2**32)
    return rdfvalue.SessionID(flow_name=id_str)

  def testSessionIDValidation(self):
    rdfvalue.SessionID(rdfvalue.RDFURN("aff4:/flows/A:12345678"))
    rdfvalue.SessionID(rdfvalue.RDFURN("aff4:/flows/A:TransferStore"))
    rdfvalue.SessionID(rdfvalue.RDFURN("aff4:/flows/DEBUG-user1:12345678"))
    rdfvalue.SessionID(rdfvalue.RDFURN("aff4:/flows/DEBUG-user1:12345678:hunt"))

  def testBadStructure(self):
    self.assertRaises(
        rdfvalue.InitializeError,
        rdfvalue.SessionID,
        rdfvalue.RDFURN("aff4:/flows/A:123456:1:"),
    )
    self.assertRaises(
        rdfvalue.InitializeError,
        rdfvalue.SessionID,
        rdfvalue.RDFURN("aff4:/flows/A:123456::"),
    )
    self.assertRaises(
        rdfvalue.InitializeError,
        rdfvalue.SessionID,
        rdfvalue.RDFURN("aff4:/flows/A:123456:"),
    )
    self.assertRaises(
        rdfvalue.InitializeError,
        rdfvalue.SessionID,
        rdfvalue.RDFURN("aff4:/flows/A:"),
    )
    self.assertRaises(
        rdfvalue.InitializeError,
        rdfvalue.SessionID,
        rdfvalue.RDFURN("aff4:/flows/:"),
    )

  def testBadQueue(self):
    self.assertRaises(
        rdfvalue.InitializeError,
        rdfvalue.SessionID,
        rdfvalue.RDFURN("aff4:/flows/A%b:12345678"),
    )

  def testBadFlowID(self):
    self.assertRaises(
        rdfvalue.InitializeError,
        rdfvalue.SessionID,
        rdfvalue.RDFURN("aff4:/flows/A:1234567G%sdf"),
    )


def main(argv):
  # Run the full test suite
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""Configuration parameters for the configuration subsystem."""


from grr_response_core.lib import config_lib

config_lib.DEFINE_string("Config.prefix",
                         "%(grr_response_core@grr-response-core|resource)",
                         "Prefix directory for general file storage.")

config_lib.DEFINE_string("Config.directory",
                         "%(install_data/etc@grr-response-core|resource)",
                         "Directory for grr server config files.")

config_lib.DEFINE_string("Config.writeback",
                         "%(Config.directory)/server.local.yaml",
                         "Location for writing back the configuration.")

config_lib.DEFINE_string(
    "ConfigUpdater.old_config", None,
    "Path to a previous config file, imported during"
    " config_updater.Initialize.")



#!/usr/bin/env python
"""WMI RDF values."""

from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import sysinfo_pb2


class WMIActiveScriptEventConsumer(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.WMIActiveScriptEventConsumer


class WMICommandLineEventConsumer(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.WMICommandLineEventConsumer



#!/usr/bin/env python
"""A module with RDF value wrappers for timeline protobufs."""

from collections.abc import Iterable, Iterator
import os

from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_core.lib.util import gzchunked
from grr_response_core.lib.util import statx
from grr_response_proto import timeline_pb2


class TimelineArgs(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the timeline arguments message."""

  protobuf = timeline_pb2.TimelineArgs
  rdf_deps = []


class TimelineResult(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the timeline result message."""

  protobuf = timeline_pb2.TimelineResult
  rdf_deps = []


class TimelineEntry(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the timeline entry message."""

  protobuf = timeline_pb2.TimelineEntry
  rdf_deps = []

  @classmethod
  def FromStat(cls, path: bytes, stat: os.stat_result) -> "TimelineEntry":
    entry = cls()
    entry.path = path

    entry.mode = stat.st_mode
    entry.size = stat.st_size

    entry.dev = stat.st_dev
    entry.ino = stat.st_ino

    entry.uid = stat.st_uid
    entry.gid = stat.st_gid

    entry.atime_ns = stat.st_atime_ns
    entry.mtime_ns = stat.st_mtime_ns
    entry.ctime_ns = stat.st_ctime_ns

    return entry

  @classmethod
  def FromStatx(cls, path: bytes, stat: statx.Result) -> "TimelineEntry":
    entry = cls()
    entry.path = path

    entry.mode = stat.mode
    entry.size = stat.size

    entry.dev = stat.dev
    entry.ino = stat.ino

    entry.uid = stat.uid
    entry.gid = stat.gid

    entry.attributes = stat.attributes

    entry.atime_ns = stat.atime_ns
    entry.btime_ns = stat.btime_ns
    entry.mtime_ns = stat.mtime_ns
    entry.ctime_ns = stat.ctime_ns

    return entry


def SerializeTimelineEntryStream(
    entries: Iterable[timeline_pb2.TimelineEntry],
) -> Iterator[bytes]:
  return gzchunked.Serialize(entry.SerializeToString() for entry in entries)


def DeserializeTimelineEntryStream(
    entries: Iterator[bytes],
) -> Iterator[timeline_pb2.TimelineEntry]:
  for entry in gzchunked.Deserialize(entries):
    parsed_entry = timeline_pb2.TimelineEntry()
    parsed_entry.ParseFromString(entry)
    yield parsed_entry


class TimelineProgress(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the timeline progress message."""

  protobuf = timeline_pb2.TimelineProgress
  rdf_deps = []



#!/usr/bin/env python
from absl import app

from grr_response_core.lib.rdfvalues import containers as rdf_containers
from grr.test_lib import test_lib


class ContainersTest(test_lib.GRRBaseTest):

  def testListContainersRequest(self):
    request = rdf_containers.ListContainersRequest()
    self.assertFalse(request.inspect_hostroot)
    request = rdf_containers.ListContainersRequest(inspect_hostroot=True)
    self.assertTrue(request.inspect_hostroot)


def main(argv):
  # Run the full test suite
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""The base classes for RDFValue tests."""

from grr_response_core.lib import serialization
from grr_response_core.lib.rdfvalues import structs as rdf_structs

# pylint:mode=test


class RDFValueTestMixin:
  """The base class for testing RDFValue implementations."""

  # This should be overridden by the RDFValue class we want to test.
  rdfvalue_class = lambda *args, **kw: None

  __abstract = True  # Do not register this class so pylint: disable=g-bad-name

  def GenerateSample(self, number=0):
    """Create a pre-populated instance of the RDFValue.

    Args:
      number: A sample number. Derived classes should return a different sample
        for each number.
    """
    _ = number
    return self.rdfvalue_class()

  def CheckRDFValue(self, value, sample):
    """Check that the rdfproto is the same as the sample."""
    self.assertIsInstance(sample, self.rdfvalue_class)
    self.assertIsInstance(value, self.rdfvalue_class)

    self.assertEqual(value, sample)

  def testComparisons(self):
    """Checks that object comparisons work."""
    sample1 = self.GenerateSample(1)

    # pylint: disable=g-generic-assert
    self.assertTrue(sample1 == self.GenerateSample(1))
    self.assertFalse(sample1 == self.GenerateSample(2))
    self.assertTrue(sample1 != self.GenerateSample(2))
    self.assertFalse(sample1 != self.GenerateSample(1))
    # pylint: enable=g-generic-assert

  def testHashability(self):
    """RDFValue instances need to act as keys in a dict."""
    sample1 = self.GenerateSample(1)

    if isinstance(sample1, rdf_structs.RDFStruct):
      self.skipTest("Hashing is unsupported.")

    # Different instances with the same value need to hash to the same.
    self.assertEqual(hash(sample1), hash(self.GenerateSample(1)))
    self.assertNotEqual(hash(sample1), hash(self.GenerateSample(2)))

  def testInitialization(self):
    """Check that we can use an empty initializer.

    RDFValues are created in many different ways, sometimes in stages by
    gradually populating fields. The only time you can be sure the user has
    finished creating a proto is when it is serialized. This means strong
    validation that requires all fields populated can't be done in init, but
    should be done in SerializeToBytes.
    """
    self.rdfvalue_class()

    # Initialize from another instance.
    sample = self.GenerateSample()

    self.CheckRDFValue(self.rdfvalue_class(sample), sample)

  def testSerialization(self, sample=None):
    """Make sure the RDFValue instance can be serialized."""
    if sample is None:
      sample = self.GenerateSample()

    # Serializing to a string must produce a string.
    serialized = serialization.ToBytes(sample)
    self.assertIsInstance(serialized, bytes)

    # Ensure we can parse it again.

    rdfvalue_object = serialization.FromBytes(self.rdfvalue_class, serialized)
    self.CheckRDFValue(rdfvalue_object, sample)

    # Serializing to data store must produce something the data store can
    # handle.
    serialized = serialization.ToWireFormat(sample)
    protobuf_type = serialization.GetProtobufType(type(sample))

    if protobuf_type == "bytes":
      self.assertIsInstance(serialized, bytes)
    elif protobuf_type == "string":
      self.assertIsInstance(serialized, str)
    elif protobuf_type in ["unsigned_integer", "integer"]:
      self.assertIsInstance(serialized, int)
    else:
      self.fail("%s has no valid protobuf_type" % self.rdfvalue_class)

    # Ensure we can parse it again.
    rdfvalue_object = serialization.FromWireFormat(
        self.rdfvalue_class, serialized
    )
    self.CheckRDFValue(rdfvalue_object, sample)


class RDFProtoTestMixin(RDFValueTestMixin):
  """A harness for testing RDFProto implementations."""

  __abstract = True  # Do not register this class so pylint: disable=g-bad-name

  def testInitializationEx(self):
    """Check we can initialize from additional parts."""
    sample = self.GenerateSample()

    # RDFProto can be initialized from a serialized protobuf.
    serialized = sample.SerializeToBytes()
    rdfvalue_sample = self.rdfvalue_class.FromSerializedBytes(serialized)
    self.CheckRDFValue(rdfvalue_sample, sample)

    # RDFProto can be initialized from another RDFProto.
    new_rdfvalue_sample = self.rdfvalue_class(rdfvalue_sample)
    self.CheckRDFValue(new_rdfvalue_sample, rdfvalue_sample)



#!/usr/bin/env python
"""Tests for utility classes."""

import datetime
import sys
import unittest

from absl import app
from absl.testing import absltest

from google.protobuf import timestamp_pb2
from grr_response_core.lib import rdfvalue
from grr.test_lib import test_lib

long_string = (
    "迎欢迎\n"
    "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Morbi luctus "
    "ex sed dictum volutpat. Integer maximus, mauris at tincidunt iaculis, "
    "felis magna scelerisque ex, in scelerisque est odio non nunc. "
    "Suspendisse et lobortis augue. Donec faucibus tempor massa, sed dapibus"
    " erat iaculis ut. Vestibulum eu elementum nulla. Nullam scelerisque "
    "hendrerit lorem. Integer vitae semper metus. Suspendisse accumsan "
    "dictum felis. Etiam viverra, felis sed ullamcorper vehicula, libero "
    "nisl tempus dui, a porta lacus erat et erat. Morbi mattis elementum "
    "efficitur. Pellentesque aliquam placerat mauris non accumsan."
)


class RDFValueTest(absltest.TestCase):
  """RDFValue tests."""

  def testStr(self):
    """Test RDFValue.__str__."""
    self.assertEqual(str(rdfvalue.RDFInteger(1)), "1")
    self.assertEqual(str(rdfvalue.RDFString(long_string)), long_string)

  # TODO(hanuszczak): Current implementation of `repr` for RDF values is broken
  # and not in line with Python guidelines. For example, `repr` should be
  # unambiguous whereas current implementation will trim long representations
  # with `...`. Moreover, the representation for most types is questionable at
  # best.
  #
  # The implementation should be fixed and proper tests should be written.


class RDFBytesTest(absltest.TestCase):

  def testFromHumanReadable(self):
    string = "zażółć gęślą jaźń"

    result = rdfvalue.RDFBytes.FromHumanReadable(string)
    expected = rdfvalue.RDFBytes.FromSerializedBytes(string.encode("utf-8"))
    self.assertEqual(result, expected)


class RDFStringTest(absltest.TestCase):

  def testFromHumanReadable(self):
    string = "pchnąć w tę łódź jeża lub ośm skrzyń fig"

    result = rdfvalue.RDFString.FromHumanReadable(string)
    self.assertEqual(str(result), string)

  def testEqualWithBytes(self):
    self.assertEqual(rdfvalue.RDFString("foo"), b"foo")
    self.assertNotEqual(rdfvalue.RDFString("foo"), b"\x80\x81\x82")

  def testLessThanWithBytes(self):
    self.assertLess(rdfvalue.RDFString("abc"), b"def")
    self.assertGreater(rdfvalue.RDFString("xyz"), b"ghi")
    self.assertLess(rdfvalue.RDFString("012"), b"\x80\x81\x81")

  # TODO: Python on Windows ships with UCS-2 by default, which does
  # not properly support unicode.
  @unittest.skipIf(
      sys.maxunicode <= 65535,
      "Your Python installation does not properly support Unicode (likely: "
      "Python with no UCS4 support on Windows.",
  )
  def testLenOfEmoji(self):
    self.assertLen(rdfvalue.RDFString("🚀🚀"), 2)


class RDFIntegerTest(absltest.TestCase):

  def testFromHumanReadable(self):
    result = rdfvalue.RDFInteger.FromHumanReadable("42")
    self.assertEqual(result, rdfvalue.RDFInteger(42))

  def testFromHumanReadablePositive(self):
    result = rdfvalue.RDFInteger.FromHumanReadable("+108")
    self.assertEqual(result, rdfvalue.RDFInteger(108))

  def testFromHumanReadableNegative(self):
    result = rdfvalue.RDFInteger.FromHumanReadable("-1337")
    self.assertEqual(result, rdfvalue.RDFInteger(-1337))

  def testFromHumanReadableZero(self):
    result = rdfvalue.RDFInteger.FromHumanReadable("0")
    self.assertEqual(result, rdfvalue.RDFInteger(0))

  def testFromHumanReadableRaisesOnNonInteger(self):
    with self.assertRaises(ValueError):
      rdfvalue.RDFInteger.FromHumanReadable("12.3")

  def testFromHumanReadableRaisesOnNonDecimal(self):
    with self.assertRaises(ValueError):
      rdfvalue.RDFInteger.FromHumanReadable("12A")


class RDFDateTimeTest(absltest.TestCase):

  def testLerpMiddle(self):
    start_time = rdfvalue.RDFDatetime.FromHumanReadable("2010-01-01")
    end_time = start_time + rdfvalue.Duration.From(10, rdfvalue.DAYS)
    lerped_time = rdfvalue.RDFDatetime.Lerp(
        0.5, start_time=start_time, end_time=end_time
    )
    self.assertEqual(
        lerped_time, start_time + rdfvalue.Duration.From(5, rdfvalue.DAYS)
    )

  def testLerpZero(self):
    start_time = rdfvalue.RDFDatetime.FromHumanReadable("2000-01-01")
    end_time = rdfvalue.RDFDatetime.FromHumanReadable("2020-01-01")
    lerped_time = rdfvalue.RDFDatetime.Lerp(
        0.0, start_time=start_time, end_time=end_time
    )
    self.assertEqual(lerped_time, start_time)

  def testLerpOne(self):
    start_time = rdfvalue.RDFDatetime.FromHumanReadable("2000-01-01")
    end_time = rdfvalue.RDFDatetime.FromHumanReadable("2020-01-01")
    lerped_time = rdfvalue.RDFDatetime.Lerp(
        1.0, start_time=start_time, end_time=end_time
    )
    self.assertEqual(lerped_time, end_time)

  def testLerpQuarter(self):
    start_time = rdfvalue.RDFDatetime.FromHumanReadable("2000-01-01")
    end_time = start_time + rdfvalue.Duration.From(4, rdfvalue.DAYS)
    lerped_time = rdfvalue.RDFDatetime.Lerp(
        0.25, start_time=start_time, end_time=end_time
    )
    self.assertEqual(
        lerped_time, start_time + rdfvalue.Duration.From(1, rdfvalue.DAYS)
    )

  def testLerpRaisesTypeErrorIfTimesAreNotRDFDatetime(self):
    now = rdfvalue.RDFDatetime.Now()

    with self.assertRaisesRegex(TypeError, "non-datetime"):
      rdfvalue.RDFDatetime.Lerp(0.0, start_time=10, end_time=now)

    with self.assertRaisesRegex(TypeError, "non-datetime"):
      rdfvalue.RDFDatetime.Lerp(
          0.0, start_time=now, end_time=rdfvalue.Duration.From(1, rdfvalue.DAYS)
      )

  def testLerpRaisesValueErrorIfProgressIsNotNormalized(self):
    start_time = rdfvalue.RDFDatetime.FromHumanReadable("2010-01-01")
    end_time = rdfvalue.RDFDatetime.FromHumanReadable("2011-01-01")

    with self.assertRaises(ValueError):
      rdfvalue.RDFDatetime.Lerp(1.5, start_time=start_time, end_time=end_time)

    with self.assertRaises(ValueError):
      rdfvalue.RDFDatetime.Lerp(-0.5, start_time=start_time, end_time=end_time)

  def testFloorToMinutes(self):
    dt = rdfvalue.RDFDatetime.FromHumanReadable("2011-11-11 12:34:56")
    expected = rdfvalue.RDFDatetime.FromHumanReadable("2011-11-11 12:34")
    self.assertEqual(
        dt.Floor(rdfvalue.Duration.From(60, rdfvalue.SECONDS)), expected
    )

  def testFloorToHours(self):
    dt = rdfvalue.RDFDatetime.FromHumanReadable("2011-11-11 12:34")
    expected = rdfvalue.RDFDatetime.FromHumanReadable("2011-11-11 12:00")
    self.assertEqual(
        dt.Floor(rdfvalue.Duration.From(1, rdfvalue.HOURS)), expected
    )

  def testFloorToDays(self):
    dt = rdfvalue.RDFDatetime.FromHumanReadable("2011-11-11 12:34")
    expected = rdfvalue.RDFDatetime.FromHumanReadable("2011-11-11")
    self.assertEqual(
        dt.Floor(rdfvalue.Duration.From(1, rdfvalue.DAYS)), expected
    )

  def testFloorExact(self):
    dt = rdfvalue.RDFDatetime.FromHumanReadable("2011-11-11 12:34:56")
    self.assertEqual(dt.Floor(rdfvalue.Duration.From(1, rdfvalue.SECONDS)), dt)

  def testAsDatetime_Epoch(self):
    py_datetime = rdfvalue.RDFDatetime(0).AsDatetime()

    self.assertIsNone(py_datetime.tzinfo)
    self.assertEqual(py_datetime.year, 1970)
    self.assertEqual(py_datetime.month, 1)
    self.assertEqual(py_datetime.day, 1)
    self.assertEqual(py_datetime.hour, 0)
    self.assertEqual(py_datetime.minute, 0)

  def testAsDatetime_Normal(self):
    rdf_datetime = rdfvalue.RDFDatetime.FromHumanReadable("1992-09-06 08:00")
    py_datetime = rdf_datetime.AsDatetime()

    self.assertIsNone(py_datetime.tzinfo)
    self.assertEqual(py_datetime.year, 1992)
    self.assertEqual(py_datetime.month, 9)
    self.assertEqual(py_datetime.day, 6)
    self.assertEqual(py_datetime.hour, 8)
    self.assertEqual(py_datetime.minute, 0)

  def testAsDatetimeUTC_Epoch(self):
    py_datetime = rdfvalue.RDFDatetime(0).AsDatetimeUTC()

    self.assertEqual(py_datetime.tzinfo, datetime.timezone.utc)
    self.assertEqual(py_datetime.year, 1970)
    self.assertEqual(py_datetime.month, 1)
    self.assertEqual(py_datetime.day, 1)
    self.assertEqual(py_datetime.hour, 0)
    self.assertEqual(py_datetime.minute, 0)

  def testAsDatetimeUTC_Normal(self):
    rdf_datetime = rdfvalue.RDFDatetime.FromHumanReadable("1992-09-06 08:00")
    py_datetime = rdf_datetime.AsDatetimeUTC()

    self.assertEqual(py_datetime.tzinfo, datetime.timezone.utc)
    self.assertEqual(py_datetime.year, 1992)
    self.assertEqual(py_datetime.month, 9)
    self.assertEqual(py_datetime.day, 6)
    self.assertEqual(py_datetime.hour, 8)
    self.assertEqual(py_datetime.minute, 0)

  def testFromTimestampProto_Now(self):
    pre_datetime = rdfvalue.RDFDatetime.Now()

    timestamp = timestamp_pb2.Timestamp()
    timestamp.GetCurrentTime()

    post_datetime = rdfvalue.RDFDatetime.Now()

    timestamp_datetime = rdfvalue.RDFDatetime.FromProtoTimestamp(timestamp)

    # Note that Protocol Buffers `Timestamp` uses nanosecond precision whereas
    # RDF `Datetime` uses microsecond precision. Thus, there is possibility that
    # during comparison we hit rounding errors, so we offset check bounds by one
    # 1 μs to account for that.
    self.assertBetween(
        timestamp_datetime,
        pre_datetime - rdfvalue.Duration.From(1, rdfvalue.MICROSECONDS),
        post_datetime + rdfvalue.Duration.From(1, rdfvalue.MICROSECONDS),
    )

  def testFromTimestampProto_Seconds(self):
    timestamp = timestamp_pb2.Timestamp(seconds=1337)

    timestamp_datetime = rdfvalue.RDFDatetime.FromProtoTimestamp(timestamp)
    self.assertEqual(timestamp_datetime.AsSecondsSinceEpoch(), 1337)

  def testFromTimestampProto_Nanos(self):
    timestamp = timestamp_pb2.Timestamp(nanos=7654321)

    timestamp_datetime = rdfvalue.RDFDatetime.FromProtoTimestamp(timestamp)
    self.assertEqual(timestamp_datetime.AsMicrosecondsSinceEpoch(), 7654)

  def testAsProtoTimestamp_Invert(self):
    timestamp = timestamp_pb2.Timestamp()
    timestamp.GetCurrentTime()

    timestamp_result = (
        rdfvalue.RDFDatetime.FromProtoTimestamp(timestamp)
    ).AsProtoTimestamp()

    # Seconds retain the precision, should be the same.
    self.assertEqual(timestamp.seconds, timestamp_result.seconds)

    # Nanos are equal only up to the microsecond precision.
    self.assertBetween(
        timestamp.nanos,
        timestamp_result.nanos - 1000,
        timestamp_result.nanos + 1000,
    )

  def testAsProtoTimestamp_Seconds(self):
    epoch = rdfvalue.RDFDatetime(0)

    timestamp = (
        epoch + rdfvalue.Duration.From(rdfvalue.SECONDS, 1337)
    ).AsProtoTimestamp()
    self.assertEqual(timestamp.seconds, 1337)
    self.assertEqual(timestamp.nanos, 0)

  def testAsProtoTimestamp_Nanos(self):
    epoch = rdfvalue.RDFDatetime(0)

    timestamp = (
        epoch + rdfvalue.Duration.From(rdfvalue.MICROSECONDS, 654321)
    ).AsProtoTimestamp()
    self.assertEqual(timestamp.seconds, 0)
    self.assertEqual(timestamp.nanos, 654321000)


class RDFDatetimeSecondsTest(absltest.TestCase):

  def testFromDatetime_withMicroSeconds(self):
    dt_with_micros = datetime.datetime(2000, 1, 1, microsecond=5000)
    dt = datetime.datetime(2000, 1, 1)
    self.assertEqual(
        rdfvalue.RDFDatetimeSeconds.FromDatetime(dt_with_micros),
        rdfvalue.RDFDatetimeSeconds.FromDatetime(dt),
    )

  def testBug122716179(self):
    d = rdfvalue.RDFDatetimeSeconds.FromSecondsSinceEpoch(1)
    self.assertEqual(d.AsMicrosecondsSinceEpoch(), 1000000)
    diff = rdfvalue.RDFDatetimeSeconds(10) - rdfvalue.Duration("3s")
    self.assertEqual(diff.AsMicrosecondsSinceEpoch(), 7000000)


class DurationSecondsTest(absltest.TestCase):

  def testPublicAttributes(self):
    duration = rdfvalue.DurationSeconds.FromHumanReadable("1h")
    self.assertEqual(duration.ToInt(rdfvalue.SECONDS), 3600)
    self.assertEqual(duration.ToInt(rdfvalue.MILLISECONDS), 3600 * 1000)
    self.assertEqual(duration.microseconds, 3600 * 1000 * 1000)

  def testFromDays(self):
    self.assertEqual(
        rdfvalue.DurationSeconds.From(2, rdfvalue.DAYS),
        rdfvalue.DurationSeconds.FromHumanReadable("2d"),
    )
    self.assertEqual(
        rdfvalue.DurationSeconds.From(31, rdfvalue.DAYS),
        rdfvalue.DurationSeconds.FromHumanReadable("31d"),
    )

  def testFromHours(self):
    self.assertEqual(
        rdfvalue.DurationSeconds.From(48, rdfvalue.HOURS),
        rdfvalue.DurationSeconds.FromHumanReadable("48h"),
    )
    self.assertEqual(
        rdfvalue.DurationSeconds.From(24, rdfvalue.HOURS),
        rdfvalue.DurationSeconds.FromHumanReadable("24h"),
    )

  def testFromSeconds(self):
    self.assertEqual(
        rdfvalue.DurationSeconds.From(1337, rdfvalue.SECONDS).ToInt(
            rdfvalue.SECONDS
        ),
        1337,
    )

  def testFromMicroseconds(self):
    duration = rdfvalue.DurationSeconds.From(3000000, rdfvalue.MICROSECONDS)
    self.assertEqual(duration.microseconds, 3000000)
    self.assertEqual(duration.ToInt(rdfvalue.SECONDS), 3)

  def testFloatConstructorRaises(self):
    with self.assertRaises(TypeError):
      rdfvalue.DurationSeconds(3.14)

  def testSerializeToBytes(self):
    self.assertEqual(
        b"0",
        rdfvalue.DurationSeconds.From(0, rdfvalue.WEEKS).SerializeToBytes(),
    )
    self.assertEqual(
        b"1",
        rdfvalue.DurationSeconds.From(1, rdfvalue.SECONDS).SerializeToBytes(),
    )
    self.assertEqual(
        b"2",
        rdfvalue.DurationSeconds.From(2, rdfvalue.SECONDS).SerializeToBytes(),
    )
    self.assertEqual(
        b"999",
        rdfvalue.DurationSeconds.From(999, rdfvalue.SECONDS).SerializeToBytes(),
    )
    self.assertEqual(
        b"1000",
        rdfvalue.DurationSeconds.From(
            1000, rdfvalue.SECONDS
        ).SerializeToBytes(),
    )

  def testFromWireFormat(self):
    for i in [0, 7, 1337]:
      val = rdfvalue.DurationSeconds.FromWireFormat(i)
      self.assertEqual(i, val.ToInt(rdfvalue.SECONDS))

      val2 = rdfvalue.DurationSeconds.FromWireFormat(
          val.SerializeToWireFormat()
      )
      self.assertEqual(val, val2)


MAX_UINT64 = 18446744073709551615


class DurationTest(absltest.TestCase):

  def testInitializationFromMicroseconds(self):
    for i in [0, 1, 7, 60, 1337, MAX_UINT64]:
      val = rdfvalue.Duration.From(i, rdfvalue.MICROSECONDS)
      self.assertEqual(i, val.microseconds)
      self.assertEqual(
          val, rdfvalue.Duration.FromHumanReadable("{} us".format(i))
      )
      self.assertEqual(val, rdfvalue.Duration(i))

  def testInitializationFromMilliseconds(self):
    for i in [0, 1, 7, 60, 1337, MAX_UINT64 // 1000]:
      val = rdfvalue.Duration.From(i, rdfvalue.MILLISECONDS)
      self.assertEqual(i * 1000, val.microseconds)
      self.assertEqual(
          val, rdfvalue.Duration.FromHumanReadable("{} ms".format(i))
      )

  def testInitializationFromSeconds(self):
    for i in [0, 1, 7, 60, 1337, MAX_UINT64 // 1000000]:
      val = rdfvalue.Duration.From(i, rdfvalue.SECONDS)
      self.assertEqual(i * 1000000, val.microseconds)
      self.assertEqual(
          val, rdfvalue.Duration.FromHumanReadable("{} s".format(i))
      )

  def testInitializationFromMinutes(self):
    for i in [0, 1, 7, 60, 1337, MAX_UINT64 // 60000000]:
      val = rdfvalue.Duration.From(i, rdfvalue.MINUTES)
      self.assertEqual(i * 60000000, val.microseconds)
      self.assertEqual(
          val, rdfvalue.Duration.FromHumanReadable("{} m".format(i))
      )

  def testInitializationFromHours(self):
    for i in [0, 1, 7, 60, 1337, MAX_UINT64 // 3600000000]:
      val = rdfvalue.Duration.From(i, rdfvalue.HOURS)
      self.assertEqual(i * 3600000000, val.microseconds)
      self.assertEqual(
          val, rdfvalue.Duration.FromHumanReadable("{} h".format(i))
      )

  def testInitializationFromDays(self):
    for i in [0, 1, 7, 60, 1337, MAX_UINT64 // 86400000000]:
      val = rdfvalue.Duration.From(i, rdfvalue.DAYS)
      self.assertEqual(i * 86400000000, val.microseconds)
      self.assertEqual(
          val, rdfvalue.Duration.FromHumanReadable("{} d".format(i))
      )

  def testInitializationFromWeeks(self):
    for i in [0, 1, 7, 60, 1337, MAX_UINT64 // 604800000000]:
      val = rdfvalue.Duration.From(i, rdfvalue.WEEKS)
      self.assertEqual(i * 604800000000, val.microseconds)
      self.assertEqual(
          val, rdfvalue.Duration.FromHumanReadable("{} w".format(i))
      )

  def testConversionToInt(self):
    for i in [0, 1, 7, 60, 1337, 12345, 123456, 1234567, MAX_UINT64]:
      val = rdfvalue.Duration.From(i, rdfvalue.MICROSECONDS)
      self.assertEqual(val.ToInt(rdfvalue.MICROSECONDS), i)
      self.assertEqual(val.ToInt(rdfvalue.MILLISECONDS), i // 1000)
      self.assertEqual(val.ToInt(rdfvalue.SECONDS), i // (1000 * 1000))
      self.assertEqual(val.ToInt(rdfvalue.MINUTES), i // (60 * 1000 * 1000))
      self.assertEqual(val.ToInt(rdfvalue.HOURS), i // (60 * 60 * 1000 * 1000))
      self.assertEqual(
          val.ToInt(rdfvalue.DAYS), i // (24 * 60 * 60 * 1000 * 1000)
      )
      self.assertEqual(
          val.ToInt(rdfvalue.WEEKS), i // (7 * 24 * 60 * 60 * 1000 * 1000)
      )

  def testConversionToFractional(self):
    for i in [0, 1, 7, 60, 1337, 12345, 123456, 1234567, MAX_UINT64]:
      val = rdfvalue.Duration.From(i, rdfvalue.MICROSECONDS)
      self.assertAlmostEqual(val.ToFractional(rdfvalue.MICROSECONDS), i)
      self.assertAlmostEqual(val.ToFractional(rdfvalue.MILLISECONDS), i / 1000)
      self.assertAlmostEqual(
          val.ToFractional(rdfvalue.SECONDS), i / (1000 * 1000)
      )
      self.assertAlmostEqual(
          val.ToFractional(rdfvalue.MINUTES), i / (60 * 1000 * 1000)
      )
      self.assertAlmostEqual(
          val.ToFractional(rdfvalue.HOURS), i / (60 * 60 * 1000 * 1000)
      )
      self.assertAlmostEqual(
          val.ToFractional(rdfvalue.DAYS), i / (24 * 60 * 60 * 1000 * 1000)
      )
      self.assertAlmostEqual(
          val.ToFractional(rdfvalue.WEEKS), i / (7 * 24 * 60 * 60 * 1000 * 1000)
      )

  def testStringDeserialization(self):
    for i in [0, 1, 7, 60, 1337, 12345, 123456, 1234567, MAX_UINT64]:
      val = rdfvalue.Duration.From(i, rdfvalue.MICROSECONDS)
      self.assertEqual(
          rdfvalue.Duration.FromSerializedBytes(val.SerializeToBytes()), val
      )

  def testHumanReadableStringSerialization(self):
    self.assertEqual("0 us", str(rdfvalue.Duration.From(0, rdfvalue.WEEKS)))
    self.assertEqual(
        "1 us", str(rdfvalue.Duration.From(1, rdfvalue.MICROSECONDS))
    )
    self.assertEqual(
        "2 us", str(rdfvalue.Duration.From(2, rdfvalue.MICROSECONDS))
    )
    self.assertEqual(
        "999 us", str(rdfvalue.Duration.From(999, rdfvalue.MICROSECONDS))
    )
    self.assertEqual(
        "1 ms", str(rdfvalue.Duration.From(1000, rdfvalue.MICROSECONDS))
    )
    self.assertEqual(
        "1 ms", str(rdfvalue.Duration.From(1, rdfvalue.MILLISECONDS))
    )
    self.assertEqual(
        "{} us".format(MAX_UINT64),
        str(rdfvalue.Duration.From(MAX_UINT64, rdfvalue.MICROSECONDS)),
    )
    self.assertEqual("3 s", str(rdfvalue.Duration.From(3, rdfvalue.SECONDS)))
    self.assertEqual("3 m", str(rdfvalue.Duration.From(3, rdfvalue.MINUTES)))
    self.assertEqual("3 h", str(rdfvalue.Duration.From(3, rdfvalue.HOURS)))
    self.assertEqual("3 d", str(rdfvalue.Duration.From(3, rdfvalue.DAYS)))
    self.assertEqual("3 w", str(rdfvalue.Duration.From(21, rdfvalue.DAYS)))

  def testSerializeToBytes(self):
    self.assertEqual(
        b"0", rdfvalue.Duration.From(0, rdfvalue.WEEKS).SerializeToBytes()
    )
    self.assertEqual(
        b"1",
        rdfvalue.Duration.From(1, rdfvalue.MICROSECONDS).SerializeToBytes(),
    )
    self.assertEqual(
        b"2",
        rdfvalue.Duration.From(2, rdfvalue.MICROSECONDS).SerializeToBytes(),
    )
    self.assertEqual(
        b"999",
        rdfvalue.Duration.From(999, rdfvalue.MICROSECONDS).SerializeToBytes(),
    )
    self.assertEqual(
        b"1000",
        rdfvalue.Duration.From(1000, rdfvalue.MICROSECONDS).SerializeToBytes(),
    )
    self.assertEqual(
        str(MAX_UINT64).encode("utf-8"),
        rdfvalue.Duration.From(
            MAX_UINT64, rdfvalue.MICROSECONDS
        ).SerializeToBytes(),
    )
    self.assertEqual(
        b"3000000",
        rdfvalue.Duration.From(3, rdfvalue.SECONDS).SerializeToBytes(),
    )

  def testAdditionOfDurationsIsEqualToIntegerAddition(self):
    for a in [0, 1, 7, 60, 1337, MAX_UINT64 // 2]:
      for b in [0, 1, 7, 60, 1337, MAX_UINT64 // 2]:
        self.assertEqual(
            rdfvalue.Duration(a) + rdfvalue.Duration(b),
            rdfvalue.Duration(a + b),
        )

  def testSubtractionOfDurationsIsEqualToIntegerSubtraction(self):
    for a in [0, 1, 7, 60, 1337, MAX_UINT64]:
      for b in [0, 1, 7, 60, 1337, MAX_UINT64]:
        self.assertEqual(
            rdfvalue.Duration(a) - rdfvalue.Duration(min(a, b)),
            rdfvalue.Duration(a - min(a, b)),
        )

  def testFromWireFormat(self):
    for i in [0, 7, 1337, MAX_UINT64]:
      val = rdfvalue.Duration.FromWireFormat(i)
      self.assertEqual(i, val.microseconds)

  def testSubtractionFromDateTimeIsEqualToIntegerSubtraction(self):
    for a in [0, 1, 7, 60, 1337]:
      for b in [0, 1, 7, 60, 1337]:
        lhs = rdfvalue.RDFDatetime.FromMicrosecondsSinceEpoch(a)
        rhs = rdfvalue.Duration(min(a, b))
        result = lhs - rhs
        self.assertEqual(result.AsMicrosecondsSinceEpoch(), a - min(a, b))

  def testAdditionToDateTimeIsEqualToIntegerAddition(self):
    for a in [0, 1, 7, 60, 1337]:
      for b in [0, 1, 7, 60, 1337]:
        lhs = rdfvalue.RDFDatetime.FromMicrosecondsSinceEpoch(a)
        rhs = rdfvalue.Duration(b)
        result = lhs + rhs
        self.assertEqual(result.AsMicrosecondsSinceEpoch(), a + b)

  def testComparisonIsEqualToIntegerComparison(self):
    for a in [0, 1, 7, 60, 1337, MAX_UINT64 - 1, MAX_UINT64]:
      for b in [0, 1, 7, 60, 1337, MAX_UINT64 - 1, MAX_UINT64]:
        dur_a = rdfvalue.Duration(a)
        dur_b = rdfvalue.Duration(b)
        if a > b:
          self.assertGreater(dur_a, dur_b)
        if a >= b:
          self.assertGreaterEqual(dur_a, dur_b)
        if a == b:
          self.assertEqual(dur_a, dur_b)
        if a <= b:
          self.assertLessEqual(dur_a, dur_b)
        if a < b:
          self.assertLess(dur_a, dur_b)
        if a != b:
          self.assertNotEqual(dur_a, dur_b)

  def testAsTimedeltaMicrosecond(self):
    duration = rdfvalue.Duration.From(1, rdfvalue.MICROSECONDS)
    self.assertEqual(duration.AsTimedelta(), datetime.timedelta(microseconds=1))

  def testAsTimedeltaSeconds(self):
    duration = rdfvalue.Duration.From(42, rdfvalue.SECONDS)
    self.assertEqual(duration.AsTimedelta(), datetime.timedelta(seconds=42))

  def testAsTimedeltaDays(self):
    duration = rdfvalue.Duration.From(1337, rdfvalue.DAYS)
    self.assertEqual(duration.AsTimedelta(), datetime.timedelta(days=1337))


class DocTest(test_lib.DocTest):
  module = rdfvalue


def main(argv):
  test_lib.main(argv)


if __name__ == "__main__":
  app.run(main)



#!/usr/bin/env python
"""A module with RDF values wrapping osquery protobufs."""
from collections.abc import Iterator, Sequence

from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import osquery_pb2


class OsqueryArgs(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the `OsqueryArgs` proto."""

  protobuf = osquery_pb2.OsqueryArgs
  rdf_deps = []

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

    if not self.HasField("timeout_millis"):
      self.timeout_millis = 5 * 60 * 1000  # 5 minutes.


class OsqueryColumn(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the `OsqueryColumn` proto."""

  protobuf = osquery_pb2.OsqueryColumn
  rdf_deps = []


class OsqueryHeader(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the `OsqueryHeader` proto."""

  protobuf = osquery_pb2.OsqueryHeader
  rdf_deps = [OsqueryColumn]


class OsqueryRow(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the `OsqueryRow` proto."""

  protobuf = osquery_pb2.OsqueryRow
  rdf_deps = []


class OsqueryTable(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the `OsqueryTable` proto."""

  protobuf = osquery_pb2.OsqueryTable
  rdf_deps = [OsqueryHeader, OsqueryRow]

  def Column(self, column_name: str) -> Iterator[str]:
    """Iterates over values of a given column.

    Args:
      column_name: A name of the column to retrieve the values for.

    Yields:
      Values of the specified column.

    Raises:
      KeyError: If given column is not present in the table.
    """
    column_idx = None
    for idx, column in enumerate(self.header.columns):
      if column.name == column_name:
        column_idx = idx
        break

    if column_idx is None:
      raise KeyError("Column '{}' not found".format(column_name))

    for row in self.rows:
      yield row.values[column_idx]


class OsqueryResult(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the `OsqueryTable` proto."""

  protobuf = osquery_pb2.OsqueryResult
  rdf_deps = [OsqueryTable]

  def GetTableColumns(self) -> Iterator[str]:
    return (column.name for column in self.table.header.columns)

  def GetTableRows(self) -> Iterator[Sequence[str]]:
    return (row.values for row in self.table.rows)


class OsqueryProgress(rdf_structs.RDFProtoStruct):
  """An RDF wrapper class for the `OsqueryProgress` proto."""

  protobuf = osquery_pb2.OsqueryProgress
  rdf_deps = [OsqueryTable]



#!/usr/bin/env python
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Acceleration module for semantic protobuf parsing."""


# pylint: disable=g-importing-member
from distutils.core import Extension
from distutils.core import setup
# pylint: enable=g-importing-member

SOURCES = ["accelerated.c"]

setup(
    name="grr",
    version="0.1",
    long_description="Semantic protobufs are smart protocol buffers.",
    license="Apache 2.0",
    author="Michael Cohen",
    ext_modules=[Extension(
        "_semantic",
        SOURCES,)],)



#!/usr/bin/env python
"""Implementation of various cryptographic types."""

import binascii
import hashlib
import logging
import os

from cryptography import exceptions
from cryptography import x509
from cryptography.hazmat.backends import openssl
from cryptography.hazmat.primitives import ciphers
from cryptography.hazmat.primitives import constant_time
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import hmac
from cryptography.hazmat.primitives import padding as sym_padding
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import modes
from cryptography.hazmat.primitives.kdf import pbkdf2
from cryptography.x509 import oid

from grr_response_core.lib import config_lib
from grr_response_core.lib import rdfvalue
from grr_response_core.lib import type_info
from grr_response_core.lib import utils
from grr_response_core.lib.rdfvalues import standard as rdf_standard
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_core.lib.util import precondition
from grr_response_core.lib.util import random
from grr_response_core.lib.util import text
from grr_response_proto import jobs_pb2


class Error(Exception):
  pass


class VerificationError(Error):
  pass


class InvalidSignature(Error):
  pass


class CipherError(rdfvalue.DecodeError):
  """Raised when decryption failed."""


class Certificate(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.Certificate


class RDFX509Cert(rdfvalue.RDFPrimitive):
  """X509 certificates used to communicate with this client."""

  def __init__(self, initializer=None):
    if initializer is None:
      super().__init__(None)
    elif isinstance(initializer, RDFX509Cert):
      super().__init__(initializer._value)  # pylint: disable=protected-access
    elif isinstance(initializer, x509.Certificate):
      super().__init__(initializer)
    elif isinstance(initializer, bytes):
      try:
        value = x509.load_pem_x509_certificate(
            initializer, backend=openssl.backend
        )
      except (ValueError, TypeError) as e:
        raise rdfvalue.DecodeError(
            "Invalid certificate %s: %s" % (initializer, e)
        )
      super().__init__(value)
    else:
      raise rdfvalue.InitializeError(
          "Cannot initialize %s from %s." % (self.__class__, initializer)
      )
    if self._value is not None:
      self.GetCN()  # This can also raise if there isn't exactly one CN entry.

  def GetRawCertificate(self):
    return self._value

  def GetCN(self):
    subject = self._value.subject
    try:
      cn_attributes = subject.get_attributes_for_oid(oid.NameOID.COMMON_NAME)
      if len(cn_attributes) > 1:
        raise rdfvalue.DecodeError("Cert has more than 1 CN entries.")
      cn_attribute = cn_attributes[0]
    except IndexError:
      raise rdfvalue.DecodeError("Cert has no CN")

    return cn_attribute.value

  def GetPublicKey(self):
    return RSAPublicKey(self._value.public_key())

  def GetSerialNumber(self):
    return self._value.serial_number

  def GetIssuer(self):
    return self._value.issuer

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)
    return cls(value)

  @classmethod
  def FromHumanReadable(cls, string: str):
    precondition.AssertType(string, str)
    return cls.FromSerializedBytes(string.encode("ascii"))

  @classmethod
  def FromWireFormat(cls, value):
    precondition.AssertType(value, bytes)
    return cls.FromSerializedBytes(value)

  def SerializeToBytes(self) -> bytes:
    if self._value is None:
      return b""
    return self._value.public_bytes(encoding=serialization.Encoding.PEM)

  # TODO(user): this should return a string, since PEM format
  # base64-encodes data and thus is ascii-compatible.
  def AsPEM(self):
    return self.SerializeToBytes()

  def __str__(self) -> str:
    return self.SerializeToBytes().decode("ascii")

  def Verify(self, public_key):
    """Verifies the certificate using the given key.

    Args:
      public_key: The public key to use.

    Returns:
      True: Everything went well.

    Raises:
      VerificationError: The certificate did not verify.
    """
    # TODO(amoser): We have to do this manually for now since cryptography does
    # not yet support cert verification. There is PR 2460:
    # https://github.com/pyca/cryptography/pull/2460/files
    # that will add it, once it's in we should switch to using this.

    # Note that all times here are in UTC.
    now = rdfvalue.RDFDatetime.Now().AsDatetime()
    if now > self._value.not_valid_after:
      raise VerificationError("Certificate expired!")
    if now < self._value.not_valid_before:
      raise VerificationError("Certificate not yet valid!")

    public_key.Verify(
        self._value.tbs_certificate_bytes,
        self._value.signature,
        hash_algorithm=self._value.signature_hash_algorithm,
    )
    return True


class CertificateSigningRequest(rdfvalue.RDFPrimitive):
  """A CSR Rdfvalue."""

  def __init__(self, initializer=None, common_name=None, private_key=None):
    if isinstance(initializer, CertificateSigningRequest):
      super().__init__(initializer._value)  # pylint: disable=protected-access
    if isinstance(initializer, x509.CertificateSigningRequest):
      super().__init__(initializer)
    elif isinstance(initializer, bytes):
      value = x509.load_pem_x509_csr(initializer, backend=openssl.backend)
      super().__init__(value)
    elif common_name and private_key:
      value = (
          x509.CertificateSigningRequestBuilder()
          .subject_name(
              x509.Name([
                  x509.NameAttribute(oid.NameOID.COMMON_NAME, str(common_name))
              ])
          )
          .sign(
              private_key.GetRawPrivateKey(),
              hashes.SHA256(),
              backend=openssl.backend,
          )
      )
      super().__init__(value)
    elif initializer is not None:
      raise rdfvalue.InitializeError(
          "Cannot initialize %s from %s." % (self.__class__, initializer)
      )

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)
    return cls(value)

  @classmethod
  def FromWireFormat(cls, value):
    precondition.AssertType(value, bytes)
    return cls.FromSerializedBytes(value)

  def SerializeToBytes(self) -> bytes:
    if self._value is None:
      return b""
    return self._value.public_bytes(serialization.Encoding.PEM)

  # TODO(user): this should return a string, since PEM format
  # base64-encodes data and thus is ascii-compatible.
  def AsPEM(self):
    return self.SerializeToBytes()

  def __str__(self) -> str:
    return self.SerializeToBytes().decode("ascii")

  def GetCN(self):
    subject = self._value.subject
    try:
      cn_attributes = subject.get_attributes_for_oid(oid.NameOID.COMMON_NAME)
      if len(cn_attributes) > 1:
        raise rdfvalue.DecodeError("CSR has more than 1 CN entries.")
      cn_attribute = cn_attributes[0]
    except IndexError:
      raise rdfvalue.DecodeError("CSR has no CN")

    return cn_attribute.value

  def GetPublicKey(self):
    return RSAPublicKey(self._value.public_key())

  def Verify(self, public_key):
    public_key.Verify(
        self._value.tbs_certrequest_bytes,
        self._value.signature,
        hash_algorithm=self._value.signature_hash_algorithm,
    )
    return True


class RSAPublicKey(rdfvalue.RDFPrimitive):
  """An RSA public key."""

  def __init__(self, initializer=None):
    if isinstance(initializer, RSAPublicKey):
      initializer = initializer._value  # pylint: disable=protected-access

    if initializer is None:
      super().__init__(None)
      return

    if isinstance(initializer, rsa.RSAPublicKey):
      super().__init__(initializer)
      return

    if isinstance(initializer, str):
      initializer = initializer.encode("ascii")

    if isinstance(initializer, bytes):
      try:
        value = serialization.load_pem_public_key(
            initializer, backend=openssl.backend
        )
        super().__init__(value)
        return
      except (TypeError, ValueError, exceptions.UnsupportedAlgorithm) as e:
        raise type_info.TypeValueError("Public key invalid: %s" % e)

    raise rdfvalue.InitializeError(
        "Cannot initialize %s from %s." % (self.__class__, initializer)
    )

  def GetRawPublicKey(self):
    return self._value

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)
    return cls(value)

  @classmethod
  def FromWireFormat(cls, value):
    precondition.AssertType(value, bytes)
    return cls.FromSerializedBytes(value)

  @classmethod
  def FromHumanReadable(cls, string: str):
    precondition.AssertType(string, str)
    return cls.FromSerializedBytes(string.encode("ascii"))

  def SerializeToBytes(self) -> bytes:
    if self._value is None:
      return b""
    return self._value.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

  def GetN(self):
    return self._value.public_numbers().n

  def __str__(self) -> str:
    return self.SerializeToBytes().decode("ascii")

  # TODO(user): this should return a string, since PEM format
  # base64-encodes data and thus is ascii-compatible.
  def AsPEM(self):
    return self.SerializeToBytes()

  def KeyLen(self):
    if self._value is None:
      return 0
    return self._value.key_size

  def Encrypt(self, message):
    if self._value is None:
      raise ValueError("Can't Encrypt with empty key.")

    try:
      return self._value.encrypt(
          message,
          padding.OAEP(
              mgf=padding.MGF1(algorithm=hashes.SHA1()),
              algorithm=hashes.SHA1(),
              label=None,
          ),
      )
    except ValueError as e:
      raise CipherError(e)

  def Verify(self, message, signature, hash_algorithm=None):
    """Verifies a given message."""
    # This method accepts both PSS and PKCS1v15 padding. PSS is preferred but
    # old clients only support PKCS1v15.

    if hash_algorithm is None:
      hash_algorithm = hashes.SHA256()

    for padding_algorithm in [
        padding.PSS(
            mgf=padding.MGF1(hash_algorithm), salt_length=padding.PSS.MAX_LENGTH
        ),
        padding.PKCS1v15(),
    ]:
      try:
        self._value.verify(
            signature, message, padding_algorithm, hash_algorithm
        )
        return True

      except exceptions.InvalidSignature as e:
        last_e = e

    raise VerificationError(last_e)


class RSAPrivateKey(rdfvalue.RDFPrimitive):
  """An RSA private key."""

  def __init__(self, initializer=None, allow_prompt=None):

    if isinstance(initializer, RSAPrivateKey):
      initializer = initializer._value  # pylint: disable=protected-access

    if initializer is None:
      super().__init__(None)
      return

    if isinstance(initializer, rsa.RSAPrivateKey):
      super().__init__(initializer)
      return

    if isinstance(initializer, str):
      initializer = initializer.encode("ascii")

    if not isinstance(initializer, bytes):
      raise rdfvalue.InitializeError(
          "Cannot initialize %s from %s." % (self.__class__, initializer)
      )

    try:
      value = serialization.load_pem_private_key(
          initializer, password=None, backend=openssl.backend
      )
      super().__init__(value)
      return
    except (TypeError, ValueError, exceptions.UnsupportedAlgorithm) as e:

      if "private key is encrypted" not in str(e):
        raise type_info.TypeValueError("Private key invalid: %s" % e)

      # The private key is passphrase protected, we need to see if we are
      # allowed to ask the user.
      #
      # In the case where allow_prompt was not set at all, we use the context
      # we are in to see if it makes sense to ask.
      if allow_prompt is None:
        # TODO(user): dependency loop with
        # core/grr_response_core/grr/config/client.py.
        # pylint: disable=protected-access
        if "Commandline Context" not in config_lib._CONFIG.context:
          raise type_info.TypeValueError("Private key invalid: %s" % e)
        # pylint: enable=protected-access

      # Otherwise, if allow_prompt is False, we are explicitly told that we are
      # not supposed to ask the user.
      elif not allow_prompt:
        raise type_info.TypeValueError("Private key invalid: %s" % e)

    try:
      # The private key is encrypted and we can ask the user for the passphrase.
      password = utils.PassphraseCallback()
      value = serialization.load_pem_private_key(
          initializer, password=password, backend=openssl.backend
      )
      super().__init__(value)
    except (TypeError, ValueError, exceptions.UnsupportedAlgorithm) as e:
      raise type_info.TypeValueError("Unable to load private key: %s" % e)

  @classmethod
  def FromHumanReadable(cls, string: str):
    precondition.AssertType(string, str)
    return cls.FromSerializedBytes(string.encode("ascii"))

  def GetRawPrivateKey(self):
    return self._value

  def GetPublicKey(self):
    return RSAPublicKey(self._value.public_key())

  def Sign(self, message):
    """Sign a given message."""
    precondition.AssertType(message, bytes)

    padding_algorithm = padding.PSS(
        mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH
    )
    return self._value.sign(message, padding_algorithm, hashes.SHA256())

  def Decrypt(self, message):
    if self._value is None:
      raise ValueError("Can't Decrypt with empty key.")

    try:
      return self._value.decrypt(
          message,
          padding.OAEP(
              mgf=padding.MGF1(algorithm=hashes.SHA1()),
              algorithm=hashes.SHA1(),
              label=None,
          ),
      )
    except ValueError as e:
      raise CipherError(e)

  @classmethod
  def GenerateKey(cls, bits=2048, exponent=65537):
    key = rsa.generate_private_key(
        public_exponent=exponent, key_size=bits, backend=openssl.backend
    )
    return cls(key)

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)
    return cls(value)

  @classmethod
  def FromWireFormat(cls, value):
    precondition.AssertType(value, bytes)

  def SerializeToBytes(self) -> bytes:
    if self._value is None:
      return b""
    return self._value.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

  def __str__(self) -> str:
    digest = hashlib.sha256(self.AsPEM()).hexdigest()

    return "%s (%s)" % ((self.__class__).__name__, digest)

  # TODO(user): this should return a string, since PEM format
  # base64-encodes data and thus is ascii-compatible.
  def AsPEM(self):
    return self.SerializeToBytes()

  def AsPassphraseProtectedPEM(self, passphrase):
    if self._value is None:
      return ""
    return self._value.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.BestAvailableEncryption(passphrase),
    )

  def KeyLen(self):
    if self._value is None:
      return 0
    return self._value.key_size


# TODO(amoser): Get rid of those.
# Conserve old names for backwards compatibility.
class PEMPrivateKey(RSAPrivateKey):
  pass


class PEMPublicKey(RSAPublicKey):
  pass


class Hash(rdf_structs.RDFProtoStruct):
  """A hash object containing multiple digests."""

  protobuf = jobs_pb2.Hash
  rdf_deps = [
      rdf_standard.AuthenticodeSignedData,
      rdfvalue.HashDigest,
  ]

  __hash__ = rdfvalue.RDFValue.__hash__


class SignedBlob(rdf_structs.RDFProtoStruct):
  """A signed blob.

  The client can receive and verify a signed blob (e.g. driver or executable
  binary). Once verified, the client may execute this.
  """

  protobuf = jobs_pb2.SignedBlob

  def Verify(self, public_key):
    """Verify the data in this blob.

    Args:
      public_key: The public key to use for verification.

    Returns:
      True when verification succeeds.

    Raises:
      rdfvalue.DecodeError if the data is not suitable verified.
    """
    if self.digest_type != self.HashType.SHA256:
      raise rdfvalue.DecodeError("Unsupported digest.")
    # TODO: Remove PKCS1v15 signature type when client adopted
    # change to PSS.
    if self.signature_type not in [
        self.SignatureType.RSA_PKCS1v15,
        self.SignatureType.RSA_PSS,
    ]:
      raise rdfvalue.DecodeError("Unsupported signature type.")

    try:
      public_key.Verify(self.data, self.signature)
    except InvalidSignature as e:
      raise rdfvalue.DecodeError("Could not verify blob. Error: %s" % e)

    return True

  def Sign(self, data, signing_key, verify_key=None):
    """Use the data to sign this blob.

    Args:
      data: String containing the blob data.
      signing_key: The key to sign with.
      verify_key: Key to verify with. If None we assume the signing key also
        contains the public key.

    Returns:
      self for call chaining.
    """

    if signing_key.KeyLen() < 2048:
      logging.warning("signing key is too short.")

    self.signature = signing_key.Sign(data)
    self.signature_type = self.SignatureType.RSA_PSS

    self.digest = hashlib.sha256(data).digest()
    self.digest_type = self.HashType.SHA256
    self.data = data

    # Test we can verify before we send it off.
    if verify_key is None:
      verify_key = signing_key.GetPublicKey()

    # Verify our own data.
    self.Verify(verify_key)

    return self


class EncryptionKey(rdfvalue.RDFPrimitive):
  """Base class for encryption keys."""

  protobuf_type = "bytes"

  def __init__(self, initializer=None):
    if initializer is None:
      super().__init__(b"")
    elif isinstance(initializer, EncryptionKey):
      super().__init__(initializer.RawBytes())
    else:
      precondition.AssertType(initializer, bytes)

      if len(initializer) % 8:
        raise CipherError(
            "Invalid key length %d (%s)." % (len(initializer) * 8, initializer)
        )

      super().__init__(initializer)

    self.length = 8 * len(self._value)
    if 0 < self.length < 128:  # Check length if _value is not empty.
      raise CipherError("Key too short (%d): %s" % (self.length, initializer))

  @classmethod
  def FromWireFormat(cls, value):
    precondition.AssertType(value, bytes)
    return cls(value)

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)
    return cls(value)

  @classmethod
  def FromHumanReadable(cls, string: str):
    precondition.AssertType(string, str)
    return cls(binascii.unhexlify(string))

  def __str__(self) -> str:
    return "%s (%s)" % (self.__class__.__name__, self.AsHexDigest())

  def __len__(self) -> int:
    return len(self._value)

  def AsHexDigest(self) -> str:
    return text.Hexify(self._value)

  def SerializeToBytes(self):
    return self._value

  @classmethod
  def GenerateKey(cls, length=128):
    return cls(os.urandom(length // 8))

  @classmethod
  def GenerateRandomIV(cls, length=128):
    return cls.GenerateKey(length=length)

  def RawBytes(self):
    return self._value


class StreamingCBCEncryptor:
  """A class to stream data to a CBCCipher object."""

  def __init__(self, cipher):
    self._cipher = cipher
    self._encryptor = cipher.GetEncryptor()
    self._overflow_buffer = b""
    self._block_size = len(cipher.key)

  def Update(self, data):
    data = self._overflow_buffer + data
    overflow_count = len(data) % self._block_size
    length_to_encrypt = len(data) - overflow_count
    to_encrypt = data[:length_to_encrypt]
    self._overflow_buffer = data[length_to_encrypt:]
    return self._encryptor.update(to_encrypt)

  def Finalize(self):
    res = self._encryptor.update(self._cipher.Pad(self._overflow_buffer))
    res += self._encryptor.finalize()
    return res


class AES128CBCCipher:
  """A Cipher using AES128 in CBC mode and PKCS7 for padding."""

  algorithm = None

  def __init__(self, key, iv):
    """Init.

    Args:
      key: The key, a rdf_crypto.EncryptionKey instance.
      iv: The iv, a rdf_crypto.EncryptionKey instance.
    """
    self.key = key.RawBytes()
    self.iv = iv.RawBytes()

  def Pad(self, data):
    padder = sym_padding.PKCS7(128).padder()
    return padder.update(data) + padder.finalize()

  def UnPad(self, padded_data):
    unpadder = sym_padding.PKCS7(128).unpadder()
    return unpadder.update(padded_data) + unpadder.finalize()

  def GetEncryptor(self):
    return ciphers.Cipher(
        algorithms.AES(self.key), modes.CBC(self.iv), backend=openssl.backend
    ).encryptor()

  def Encrypt(self, data):
    """A convenience method which pads and encrypts at once."""
    encryptor = self.GetEncryptor()
    padded_data = self.Pad(data)

    try:
      return encryptor.update(padded_data) + encryptor.finalize()
    except ValueError as e:
      raise CipherError(e)

  def GetDecryptor(self):
    return ciphers.Cipher(
        algorithms.AES(self.key), modes.CBC(self.iv), backend=openssl.backend
    ).decryptor()

  def Decrypt(self, data):
    """A convenience method which pads and decrypts at once."""
    decryptor = self.GetDecryptor()

    try:
      padded_data = decryptor.update(data) + decryptor.finalize()
      return self.UnPad(padded_data)
    except ValueError as e:
      raise CipherError(e)


class SymmetricCipher(rdf_structs.RDFProtoStruct):
  """Abstract symmetric cipher operations."""

  protobuf = jobs_pb2.SymmetricCipher
  rdf_deps = [
      EncryptionKey,
  ]

  @classmethod
  def Generate(cls, algorithm):
    if algorithm != cls.Algorithm.AES128CBC:
      raise RuntimeError("Algorithm not supported.")

    return cls(
        _algorithm=algorithm,
        _key=EncryptionKey.GenerateKey(length=128),
        _iv=EncryptionKey.GenerateKey(length=128),
    )

  def _get_cipher(self):
    if self._algorithm != self.Algorithm.AES128CBC:
      raise CipherError("Unknown cipher type %s" % self._algorithm)

    return AES128CBCCipher(self._key, self._iv)

  def Encrypt(self, data):
    if self._algorithm == self.Algorithm.NONE:
      raise TypeError("Empty encryption is not allowed.")

    return self._get_cipher().Encrypt(data)

  def Decrypt(self, data):
    if self._algorithm == self.Algorithm.NONE:
      raise TypeError("Empty encryption is not allowed.")

    return self._get_cipher().Decrypt(data)


class HMAC:
  """A wrapper for the cryptography HMAC object."""

  def __init__(self, key, use_sha256=False):
    # We store the raw key from cryptography.io.
    if isinstance(key, EncryptionKey):
      key = key.RawBytes()

    self.key = key
    self._hmac = self._NewHMAC(use_sha256=use_sha256)

  def _NewHMAC(self, use_sha256=False):
    if use_sha256:
      hash_algorithm = hashes.SHA256()
    else:
      hash_algorithm = hashes.SHA1()
    return hmac.HMAC(self.key, hash_algorithm, backend=openssl.backend)

  def Update(self, data):
    self._hmac.update(data)

  def Finalize(self):
    return self._hmac.finalize()

  def HMAC(self, message, use_sha256=False):
    """Calculates the HMAC for a given message."""
    h = self._NewHMAC(use_sha256=use_sha256)
    h.update(message)
    return h.finalize()

  def Verify(self, message, signature):
    """Verifies the signature for a given message."""
    siglen = len(signature)
    if siglen == 20:
      hash_algorithm = hashes.SHA1()
    elif siglen == 32:
      hash_algorithm = hashes.SHA256()
    else:
      raise VerificationError("Invalid signature length %d." % siglen)

    h = hmac.HMAC(self.key, hash_algorithm, backend=openssl.backend)
    h.update(message)
    try:
      h.verify(signature)
      return True
    except exceptions.InvalidSignature as e:
      raise VerificationError(e)


def _CalculateHash(password: bytes, salt: bytes, iteration_count: int) -> bytes:
  kdf = pbkdf2.PBKDF2HMAC(
      algorithm=hashes.SHA256(),
      length=32,
      salt=salt,
      iterations=iteration_count,
      backend=openssl.backend,
  )
  return kdf.derive(password)


def SetPassword(proto: jobs_pb2.Password, password: str) -> None:
  """Sets the password in the proto."""

  proto.salt = b"%016x" % random.UInt64()
  proto.iteration_count = 100000

  # prevent non-descriptive 'key_material must be bytes' error later
  password_bytes = password.encode("utf-8")

  proto.hashed_pwd = _CalculateHash(
      password_bytes, proto.salt, proto.iteration_count
  )


def CheckPassword(proto: jobs_pb2.Password, password: str) -> bool:
  # prevent non-descriptive 'key_material must be bytes' error later
  password_bytes = password.encode("utf-8")

  h = _CalculateHash(password_bytes, proto.salt, proto.iteration_count)
  return constant_time.bytes_eq(h, proto.hashed_pwd)


class Password(rdf_structs.RDFProtoStruct):
  """A password stored in the database."""

  protobuf = jobs_pb2.Password



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import config as rdf_config
from grr_response_proto import config_pb2


def ToProtoAdminUIClientWarningRule(
    rdf: rdf_config.AdminUIClientWarningRule,
) -> config_pb2.AdminUIClientWarningRule:
  return rdf.AsPrimitiveProto()


def ToRDFAdminUIClientWarningRule(
    proto: config_pb2.AdminUIClientWarningRule,
) -> rdf_config.AdminUIClientWarningRule:
  return rdf_config.AdminUIClientWarningRule.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoAdminUIClientWarningsConfigOption(
    rdf: rdf_config.AdminUIClientWarningsConfigOption,
) -> config_pb2.AdminUIClientWarningsConfigOption:
  return rdf.AsPrimitiveProto()


def ToRDFAdminUIClientWarningsConfigOption(
    proto: config_pb2.AdminUIClientWarningsConfigOption,
) -> rdf_config.AdminUIClientWarningsConfigOption:
  return rdf_config.AdminUIClientWarningsConfigOption.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""This file contains cache-related utility functions used by GRR."""

from collections.abc import Callable
import functools
import logging
import threading
from typing import Any, TypeVar

from grr_response_core.lib import rdfvalue

WITH_LIMITED_CALL_FREQUENCY_PASS_THROUGH = False

_F = TypeVar("_F", bound=Callable[..., Any])

_FVoid = TypeVar("_FVoid", bound=Callable[..., None])


def WithLimitedCallFrequency(
    min_time_between_calls: rdfvalue.Duration,
) -> Callable[[_F], _F]:
  """Function call rate-limiting decorator.

  This decorator ensures that the wrapped function will be called at most
  once in min_time_between_calls time for the same set of arguments. For all
  excessive calls a previous cached return value will be returned.

  Suppose we use the decorator like this:
  @cache.WithLimitedCallFrequency(rdfvalue.Duration.From(30, rdfvalue.SECONDS))
  def Foo(id):
    ...

  If Foo(42) is called and then Foo(42) is called again within 30 seconds, then
  the second call will simply return the cached return value of the first.

  If Foo(42) is called and then Foo(43) is called within 30 seconds, the
  wrapped function will be properly called in both cases, since these Foo calls
  have different arguments sets.

  If Foo(42) is called and takes a long time to finish, and another
  Foo(42) call is done in another thread, then the latter call will wait for
  the first one to finish and then return the cached result value. I.e. the
  wrapped function will be called just once, thus keeping the guarantee of
  at most 1 run in min_time_between_calls.

  NOTE 1: this function becomes a trivial pass-through and does no caching if
  module-level WITH_LIMITED_CALL_FREQUENCY_PASS_THROUGH variable is set to
  True. This is used in testing.

  NOTE 2: all decorated functions' arguments have to be hashable.

  Args:
    min_time_between_calls: An rdfvalue.Duration specifying the minimal time to
      pass between 2 consecutive function calls with same arguments.

  Returns:
    A Python function decorator.
  """

  def Decorated(f: _F) -> _F:
    """Actual decorator implementation."""

    lock = threading.RLock()
    prev_times = {}
    prev_results = {}
    result_locks = {}
    prev_cleanup_time = rdfvalue.RDFDatetime.Now()

    def CleanUpCache(now: rdfvalue.RDFDatetime, min_time: rdfvalue.Duration):
      """Cleans up the cache from stale entries."""
      nonlocal prev_cleanup_time
      if now < prev_cleanup_time:
        logging.warning(
            "Current timestamp %s is before the previous cache cleaning time"
            " %s, hoping we're inside the test",
            now,
            prev_cleanup_time,
        )
        prev_cleanup_time = now
        return

      if (now - prev_cleanup_time) < min_time:
        return

      for k, prev_time in list(prev_times.items()):
        if prev_time > now:
          # We have a result from the future, hopefully this is a test...
          logging.warning(
              "Deleting cached function result from the future (%s > %s)",
              prev_time,
              now,
          )
          prev_times.pop(k)
          prev_results.pop(k, None)
          result_locks.pop(k, None)
        elif now - prev_time >= min_time:
          prev_times.pop(k)
          prev_results.pop(k, None)
          result_locks.pop(k, None)

      prev_cleanup_time = now

    @functools.wraps(f)
    def Fn(*args, **kwargs):
      """Wrapper around the decorated function."""

      if WITH_LIMITED_CALL_FREQUENCY_PASS_THROUGH:
        # This effectively turns off the caching.
        min_time = rdfvalue.Duration(0)
      else:
        min_time = min_time_between_calls

      key = (args, tuple(sorted(kwargs.items())))
      now = rdfvalue.RDFDatetime.Now()

      with lock:
        CleanUpCache(now, min_time)

        try:
          prev_time = prev_times[key]
          if now - prev_time < min_time:
            return prev_results[key]
        except KeyError:
          prev_time = None

        try:
          result_lock = result_locks[key]
        except KeyError:
          result_lock = threading.RLock()
          result_locks[key] = result_lock

      with result_lock:
        t = prev_times.get(key)

        if t == prev_time:
          result = f(*args, **kwargs)
          with lock:
            prev_times[key] = rdfvalue.RDFDatetime.Now()
            prev_results[key] = result

          return result
        else:
          return prev_results[key]

    def _DebugInternalState():
      return dict(
          prev_times=prev_times,
          prev_results=prev_results,
          result_locks=result_locks,
          prev_cleanup_time=prev_cleanup_time,
      )

    # This is used by the tests to ensure that the internal representation
    # behaves as expected.
    Fn._DebugInternalState = (  # pylint: disable=protected-access
        _DebugInternalState
    )

    return Fn

  return Decorated


def WithLimitedCallFrequencyWithoutReturnValue(
    min_time_between_calls: rdfvalue.Duration,
) -> Callable[[_FVoid], _FVoid]:
  """Function call rate-limiting decorator for None-returning functions.

  This decorator ensures that the wrapped function will be called at most
  once in min_time_between_calls time for the same set of arguments. Given
  that the wrapped function is not expected to return a value, all excessive
  calls will be dropped immediately, even if a parallel ongoing
  call for the same set of arguments is in progress in another thread.

  Suppose we use the decorator like this:
  @cache.WithLimitedCallFrequencyWithoutReturnValue(
      rdfvalue.Duration.From(30, rdfvalue.SECONDS))
  def Foo(id):
    ...

  If Foo(42) is called and then Foo(42) is called again within 30 seconds, then
  the second call will simply return immediately.

  If Foo(42) is called and then Foo(43) is called within 30 seconds, the
  wrapped function will be properly called in both cases, since these Foo calls
  have different arguments sets.

  If Foo(42) is called and takes a long time to finish, and another
  Foo(42) call is done in another thread, then the latter call will return
  immediately.

  NOTE 1: this function becomes a trivial pass-through and does no caching if
  module-level WITH_LIMITED_CALL_FREQUENCY_PASS_THROUGH variable is set to
  True. This is used in testing.

  NOTE 2: all decorated functions' arguments have to be hashable.

  Args:
    min_time_between_calls: An rdfvalue.Duration specifying the minimal time to
      pass between 2 consecutive function calls with same arguments.

  Returns:
    A Python function decorator.
  """

  def Decorated(f: _FVoid) -> _FVoid:
    """Actual decorator implementation."""

    lock = threading.RLock()
    prev_times = {}
    in_progress_locks = {}
    prev_cleanup_time = rdfvalue.RDFDatetime.Now()

    def CleanUpCache(now: rdfvalue.RDFDatetime, min_time: rdfvalue.Duration):
      """Cleans up the cache from stale entries."""
      nonlocal prev_cleanup_time
      if now < prev_cleanup_time:
        logging.warning(
            "Current timestamp %s is before the previous cache cleaning time"
            " %s, hoping we're inside the test",
            now,
            prev_cleanup_time,
        )
        prev_cleanup_time = now
        return

      if (now - prev_cleanup_time) < min_time:
        return

      for k, prev_time in list(prev_times.items()):
        if prev_time > now:
          # We have a result from the future, hopefully this is a test...
          logging.warning(
              "Deleting cached function result from the future (%s > %s)",
              prev_time,
              now,
          )
          prev_times.pop(k)
          in_progress_locks.pop(k, None)
        elif now - prev_time >= min_time:
          prev_times.pop(k)
          in_progress_locks.pop(k, None)

      prev_cleanup_time = now

    @functools.wraps(f)
    def Fn(*args, **kwargs):
      """Wrapper around the decorated function."""

      if WITH_LIMITED_CALL_FREQUENCY_PASS_THROUGH:
        # This effectively turns off the caching.
        min_time = rdfvalue.Duration(0)
      else:
        min_time = min_time_between_calls

      key = (args, tuple(sorted(kwargs.items())))
      now = rdfvalue.RDFDatetime.Now()

      with lock:
        CleanUpCache(now, min_time)

        try:
          prev_time = prev_times[key]
          if now - prev_time < min_time:
            return
        except KeyError:
          pass

        try:
          in_progress_lock = in_progress_locks[key]
        except KeyError:
          in_progress_lock = threading.RLock()
          in_progress_locks[key] = in_progress_lock

      if in_progress_lock.acquire(blocking=False):
        try:
          r = f(*args, **kwargs)
          assert r is None, "Wrapped function should have no return value"

          with lock:
            prev_times[key] = rdfvalue.RDFDatetime.Now()
        finally:
          in_progress_lock.release()

    def _DebugInternalState():
      return dict(
          prev_times=prev_times,
          in_progress_locks=in_progress_locks,
          prev_cleanup_time=prev_cleanup_time,
      )

    # This is used by the tests to ensure that the internal representation
    # behaves as expected.
    Fn._DebugInternalState = (  # pylint: disable=protected-access
        _DebugInternalState
    )

    return Fn

  return Decorated



#!/usr/bin/env python
"""An LL(1) lexer. This lexer is very tolerant of errors and can resync."""

import logging
import re

from grr_response_core.lib import utils
from grr_response_core.lib.util import precondition
from grr_response_core.lib.util import text


class Token:
  """A token action."""

  state_regex = None

  def __init__(self, state_regex, regex, actions, next_state, flags=re.I):
    """Constructor.

    Args:
      state_regex: If this regular expression matches the current state this
        rule is considered.
      regex: A regular expression to try and match from the current point.
      actions: A command separated list of method names in the Lexer to call.
      next_state: The next state we transition to if this Token matches.
      flags: re flags.
    """
    precondition.AssertType(regex, str)
    precondition.AssertOptionalType(state_regex, str)

    if state_regex:
      self.state_regex = re.compile(
          state_regex, re.DOTALL | re.M | re.S | re.U | flags
      )
    self.regex = re.compile(regex, re.DOTALL | re.M | re.S | re.U | flags)

    self.re_str = regex
    self.actions = []
    if actions:
      self.actions = actions.split(",")

    self.next_state = next_state

  def Action(self, lexer):
    """Method is called when the token matches."""


class Error(Exception):
  """Module exception."""


class ParseError(Error):
  """A parse error occurred."""


class Lexer:
  """A generic feed lexer."""

  # A list of Token() instances.
  tokens = []
  # Regex flags
  flags = 0

  def __init__(self, data=""):
    precondition.AssertType(data, str)
    # Set the lexer up to process a new data feed.
    self.Reset()
    # Populate internal token list with class tokens, if defined.
    self._tokens = self.tokens[:]
    # Populate the lexer with any data we got.
    self.buffer = data

  def Reset(self):
    """Reset the lexer to process a new data feed."""
    # The first state
    self.state = "INITIAL"
    self.state_stack = []

    # The buffer we are parsing now
    self.buffer = ""
    self.error = 0
    self.verbose = 0

    # The index into the buffer where we are currently pointing
    self.processed = 0
    self.processed_buffer = ""

  def NextToken(self):
    """Fetch the next token by trying to match any of the regexes in order."""
    # Nothing in the input stream - no token can match.
    if not self.buffer:
      return

    current_state = self.state
    for token in self._tokens:
      # Does the rule apply to us?
      if token.state_regex and not token.state_regex.match(current_state):
        continue

      if self.verbose:
        logging.debug(
            "%s: Trying to match %r with %r",
            self.state,
            self.buffer[:10],
            token.re_str,
        )

      # Try to match the rule
      m = token.regex.match(self.buffer)
      if not m:
        continue

      if self.verbose:
        logging.debug("%s matched %s", token.re_str, m.group(0))

      # A token matched the empty string. We can not consume the token from the
      # input stream.
      if m.end() == 0:
        raise RuntimeError("Lexer bug! Token can not match the empty string.")

      # The match consumes the data off the buffer (the handler can put it back
      # if it likes)
      self.processed_buffer += self.buffer[: m.end()]
      self.buffer = self.buffer[m.end() :]
      self.processed += m.end()

      next_state = token.next_state
      for action in token.actions:
        if self.verbose:
          logging.debug("Calling %s with %s", action, m.group(0))

        # Is there a callback to handle this action?
        cb = getattr(self, action, self.Default)

        # Allow a callback to skip other callbacks.
        try:
          possible_next_state = cb(string=m.group(0), match=m)
          if possible_next_state == "CONTINUE":
            continue
          # Override the state from the Token
          elif possible_next_state:
            next_state = possible_next_state
        except ParseError as e:
          self.Error(e)

      # Update the next state
      if next_state:
        self.state = next_state

      return token

    # Check that we are making progress - if we are too full, we assume we are
    # stuck.
    self.Error("Lexer stuck at state %s" % (self.state))
    self.processed_buffer += self.buffer[:1]
    self.buffer = self.buffer[1:]
    return "Error"

  def Feed(self, data):
    precondition.AssertType(data, str)
    self.buffer += data

  def Empty(self):
    return not self.buffer

  def Default(self, **kwarg):
    logging.debug("Default handler: %s", kwarg)

  def Error(self, message=None, weight=1):
    logging.debug("Error(%s): %s", weight, message)
    # Keep a count of errors
    self.error += weight

  def PushState(self, **_):
    """Push the current state on the state stack."""
    if self.verbose:
      logging.debug("Storing state %r", self.state)
    self.state_stack.append(self.state)

  def PopState(self, **_):
    """Pop the previous state from the stack."""
    try:
      self.state = self.state_stack.pop()
      if self.verbose:
        logging.debug("Returned state to %s", self.state)

      return self.state
    except IndexError:
      self.Error("Tried to pop the state but failed - possible recursion error")

  def PushBack(self, string="", **_):
    """Push the match back on the stream."""
    precondition.AssertType(string, str)
    self.buffer = string + self.buffer
    self.processed_buffer = self.processed_buffer[: -len(string)]

  def Close(self):
    """A convenience function to force us to parse all the data."""
    while self.NextToken():
      if not self.buffer:
        return


class Expression:
  """A class representing an expression."""

  attribute = None
  args = None
  operator = None

  # The expected number of args
  number_of_args = 1

  def __init__(self):
    self.args = []

  def SetAttribute(self, attribute):
    self.attribute = attribute

  def SetOperator(self, operator):
    self.operator = operator

  def AddArg(self, arg):
    """Adds a new arg to this expression.

    Args:
       arg: The argument to add (string).

    Returns:
      True if this arg is the last arg, False otherwise.

    Raises:
      ParseError: If there are too many args.
    """
    self.args.append(arg)
    if len(self.args) > self.number_of_args:
      raise ParseError("Too many args for this expression.")

    elif len(self.args) == self.number_of_args:
      return True

    return False

  def __str__(self):
    return "Expression: (%s) (%s) %s" % (
        self.attribute,
        self.operator,
        self.args,
    )

  def PrintTree(self, depth=""):
    return "%s %s" % (depth, self)

  def Compile(self, filter_implemention):
    """Given a filter implementation, compile this expression."""
    raise NotImplementedError(
        "%s does not implement Compile." % self.__class__.__name__
    )


class BinaryExpression(Expression):
  """An expression which takes two other expressions."""

  def __init__(self, operator="", part=None):
    self.operator = operator
    self.args = []
    if part:
      self.args.append(part)
    super().__init__()

  def __str__(self) -> str:
    return "Binary Expression: %s %s" % (
        self.operator,
        [str(x) for x in self.args],
    )

  def AddOperands(self, lhs, rhs):
    if isinstance(lhs, Expression) and isinstance(rhs, Expression):
      self.args.insert(0, lhs)
      self.args.append(rhs)
    else:
      raise ParseError(
          "Expected expression, got %s %s %s" % (lhs, self.operator, rhs)
      )

  def PrintTree(self, depth=""):
    result = "%s%s\n" % (depth, self.operator)
    for part in self.args:
      result += "%s-%s\n" % (depth, part.PrintTree(depth + "  "))

    return result

  def Compile(self, filter_implemention):
    """Compile the binary expression into a filter object."""
    operator = self.operator.lower()
    if operator == "and" or operator == "&&":
      method = "AndFilter"
    elif operator == "or" or operator == "||":
      method = "OrFilter"
    else:
      raise ParseError("Invalid binary operator %s" % operator)

    args = [x.Compile(filter_implemention) for x in self.args]
    return filter_implemention.GetFilter(method)(*args)


class IdentityExpression(Expression):
  """An Expression which always evaluates to True."""

  def Compile(self, filter_implemention):
    return filter_implemention.IdentityFilter()


class SearchParser(Lexer):
  """This parser can parse the mini query language and build an AST.

  Examples of valid syntax:
    filename contains "foo" and (size > 100k or date before "2011-10")
    date between 2011 and 2010
    files older than 1 year
  """

  expression_cls = Expression
  binary_expression_cls = BinaryExpression
  identity_expression_cls = IdentityExpression
  string = ""

  tokens = [
      # Double quoted string
      Token("STRING", '"', "PopState,StringFinish", None),
      Token("STRING", r"\\(.)", "StringEscape", None),
      Token("STRING", r"[^\\\"]+", "StringInsert", None),
      # Single quoted string
      Token("SQ_STRING", "'", "PopState,StringFinish", None),
      Token("SQ_STRING", r"\\(.)", "StringEscape", None),
      Token("SQ_STRING", r"[^\\']+", "StringInsert", None),
      # TODO(user): Implement a unary not operator.
      # The first thing we see in the initial state takes up to the ATTRIBUTE
      Token("INITIAL", r"(and|or|\&\&|\|\|)", "BinaryOperator", None),
      Token("INITIAL", r"[^\s\(\)]", "PushState,PushBack", "ATTRIBUTE"),
      Token("INITIAL", r"\(", "BracketOpen", None),
      Token("INITIAL", r"\)", "BracketClose", None),
      Token("ATTRIBUTE", r"[\w._0-9]+", "StoreAttribute", "OPERATOR"),
      Token(
          "OPERATOR", r"[a-z0-9<>=\-\+\!\^\&%]+", "StoreOperator", "ARG_LIST"
      ),
      Token("OPERATOR", "(!=|[<>=])", "StoreSpecialOperator", "ARG_LIST"),
      Token("ARG_LIST", r"[^\s'\"]+", "InsertArg", None),
      # Start a string.
      Token(".", '"', "PushState,StringStart", "STRING"),
      Token(".", "'", "PushState,StringStart", "SQ_STRING"),
      # Skip whitespace.
      Token(".", r"\s+", None, None),
  ]

  def __init__(self, data):
    # Holds expression
    self.current_expression = self.expression_cls()
    self.filter_string = data

    # The token stack
    self.stack = []
    Lexer.__init__(self, data)

  def BinaryOperator(self, string=None, **_):
    self.stack.append(self.binary_expression_cls(string))

  def BracketOpen(self, **_):
    self.stack.append("(")

  def BracketClose(self, **_):
    self.stack.append(")")

  def StringStart(self, **_):
    self.string = ""

  def StringEscape(self, string, match, **_):
    """Escape backslashes found inside a string quote.

    Backslashes followed by anything other than ['"rnbt] will just be included
    in the string.

    Args:
       string: The string that matched.
       match: The match object (m.group(1) is the escaped code)
    """
    precondition.AssertType(string, str)
    if match.group(1) in "'\"rnbt":
      self.string += text.Unescape(string)
    else:
      self.string += string

  def StringInsert(self, string="", **_):
    self.string += string

  def StringFinish(self, **_):
    if self.state == "ATTRIBUTE":
      return self.StoreAttribute(string=self.string)

    elif self.state == "ARG_LIST":
      return self.InsertArg(string=self.string)

  def StoreAttribute(self, string="", **_):
    if self.verbose:
      logging.debug("Storing attribute %r", string)

    # TODO(user): Update the expected number_of_args
    try:
      self.current_expression.SetAttribute(string)
    except AttributeError:
      raise ParseError("Invalid attribute '%s'" % string)

    return "OPERATOR"

  def StoreOperator(self, string="", **_):
    if self.verbose:
      logging.debug("Storing operator %r", string)
    self.current_expression.SetOperator(string)

  def InsertArg(self, string="", **_):
    """Insert an arg to the current expression."""
    if self.verbose:
      logging.debug("Storing Argument %s", utils.SmartUnicode(string))

    # This expression is complete
    if self.current_expression.AddArg(string):
      self.stack.append(self.current_expression)
      self.current_expression = self.expression_cls()
      return self.PopState()

  def _CombineBinaryExpressions(self, operator):
    for i in range(1, len(self.stack) - 1):
      item = self.stack[i]
      if (
          isinstance(item, BinaryExpression)
          and item.operator == operator
          and isinstance(self.stack[i - 1], Expression)
          and isinstance(self.stack[i + 1], Expression)
      ):
        lhs = self.stack[i - 1]
        rhs = self.stack[i + 1]

        item.AddOperands(lhs, rhs)
        self.stack[i - 1] = None
        self.stack[i + 1] = None

    self.stack = list(filter(None, self.stack))

  def _CombineParenthesis(self):
    for i in range(len(self.stack) - 2):
      if (
          self.stack[i] == "("
          and self.stack[i + 2] == ")"
          and isinstance(self.stack[i + 1], Expression)
      ):
        self.stack[i] = None
        self.stack[i + 2] = None

    self.stack = list(filter(None, self.stack))

  def Reduce(self):
    """Reduce the token stack into an AST."""
    # Check for sanity
    if self.state != "INITIAL":
      self.Error("Premature end of expression")

    length = len(self.stack)
    while length > 1:
      # Precedence order
      self._CombineParenthesis()
      self._CombineBinaryExpressions("and")
      self._CombineBinaryExpressions("or")

      # No change
      if len(self.stack) == length:
        break
      length = len(self.stack)

    if length != 1:
      self.Error("Illegal query expression")

    return self.stack[0]

  def Error(self, message=None, weight=1):
    raise ParseError(
        "%s in position %s: %s <----> %s )"
        % (
            utils.SmartUnicode(message),
            len(self.processed_buffer),
            self.processed_buffer,
            self.buffer,
        )
    )

  def Parse(self):
    if not self.filter_string:
      return self.identity_expression_cls()

    self.Close()
    return self.Reduce()



#!/usr/bin/env python
"""This is the GRR config management code.

This handles opening and parsing of config files.
"""

from collections import abc
import copy
import io
import logging
import os
import platform
import re
import sys
import traceback
from typing import Any, BinaryIO, Optional, cast

from absl import flags

from grr_response_core.lib import config_parser
from grr_response_core.lib import lexer
from grr_response_core.lib import package
from grr_response_core.lib import type_info
from grr_response_core.lib import utils
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_core.lib.registry import MetaclassRegistry
from grr_response_core.lib.util import precondition


# Default is set in distro_entry.py to be taken from package resource.
flags.DEFINE_string(
    "config",
    package.ResourcePath(
        "grr-response-core", "install_data/etc/grr-server.yaml"
    ),
    "Primary Configuration file to use. This is normally "
    "taken from the installed package and should rarely "
    "be specified.",
)

flags.DEFINE_list(
    "secondary_configs",
    [],
    "Secondary configuration files to load (These override "
    "previous configuration files.).",
)

flags.DEFINE_bool("config_help", False, "Print help about the configuration.")

flags.DEFINE_list("context", [], "Use these contexts for the config.")

flags.DEFINE_bool(
    "disallow_missing_config_definitions",
    False,
    "If true, we raise an error on undefined config options. "
    "This flag has an effect only on clients because it is set "
    "as True by default for server components.",
)

flags.DEFINE_multi_string(
    "parameter",
    default=[],
    help=(
        "Global override of config values. "
        "For example -p Database.implementation: MysqlDB"
    ),
    short_name="p",
)


class Error(Exception):
  """Base class for configuration exceptions."""


class ConfigFormatError(Error, type_info.TypeValueError):
  """Raised when configuration file is formatted badly."""


class ConfigWriteError(Error):
  """Raised when we failed to update the config."""


class ConfigFileNotFound(IOError, Error):
  """Raised when a config file was expected but was not found."""


class UnknownOption(Error, KeyError):
  """Raised when an unknown option was requested."""


class InterpolationError(Error):
  """Raised when a config object failed to interpolate."""


class FilterError(InterpolationError):
  """Raised when a filter fails to perform its function."""


class ConstModificationError(Error):
  """Raised when the config tries to change a constant option."""


class AlreadyInitializedError(Error):
  """Raised when an option is defined after initialization."""


class MissingConfigDefinitionError(Error):
  """Raised when a config contains an undefined config option."""


class InvalidContextError(Error):
  """Raised when an invalid context is used."""


def SetPlatformArchContext():
  """Add the running contexts to the config system."""

  # Initialize the running platform context:
  _CONFIG.AddContext("Platform:%s" % platform.system().title())

  machine = platform.uname()[4]
  if machine in ["x86_64", "AMD64", "i686"]:
    # 32 bit binaries running on AMD64 will still have a i386 arch.
    if platform.architecture()[0] == "32bit":
      arch = "i386"
    else:
      arch = "amd64"
  elif machine == "x86":
    arch = "i386"
  elif machine == "arm64":
    arch = "aarch64"
  else:
    arch = machine

  _CONFIG.AddContext("Arch:%s" % arch)


class ConfigFilter(metaclass=MetaclassRegistry):
  """A configuration filter can transform a configuration parameter."""

  name = "identity"

  # If this is set, application of the filter will not be logged - useful
  # for key material.
  sensitive_arg = False

  def Filter(self, data: str) -> str:
    precondition.AssertType(data, str)
    return data


class Literal(ConfigFilter):
  """A filter which does not interpolate."""

  name = "literal"


class Lower(ConfigFilter):
  name = "lower"

  def Filter(self, data: str) -> str:
    precondition.AssertType(data, str)
    return data.lower()


class Upper(ConfigFilter):
  name = "upper"

  def Filter(self, data: str) -> str:
    precondition.AssertType(data, str)
    return data.upper()


class Filename(ConfigFilter):
  name = "file"

  def Filter(self, data: str) -> str:
    precondition.AssertType(data, str)
    try:
      with io.open(data, "r") as fd:
        return fd.read()  # pytype: disable=bad-return-type
    except IOError as e:
      raise FilterError("%s: %s" % (data, e))


class OptionalFile(ConfigFilter):
  name = "optionalfile"

  def Filter(self, data: str) -> str:
    precondition.AssertType(data, str)
    try:
      with io.open(data, "r") as fd:
        return fd.read()  # pytype: disable=bad-return-type
    except IOError:
      return ""


class FixPathSeparator(ConfigFilter):
  """A configuration filter that fixes the path speratator."""

  name = "fixpathsep"

  def Filter(self, data: str) -> str:
    precondition.AssertType(data, str)
    if platform.system() == "Windows":
      # This will fix "X:\", and might add extra slashes to other paths, but
      # this is OK.
      return data.replace("\\", "\\\\")
    else:
      return data.replace("\\", "/")


class Env(ConfigFilter):
  """Interpolate environment variables."""

  name = "env"

  def Filter(self, data: str) -> str:
    precondition.AssertType(data, str)
    return os.environ.get(data.upper(), "")


class Expand(ConfigFilter):
  """Expands the input as a configuration parameter."""

  name = "expand"

  def Filter(self, data: str) -> str:
    precondition.AssertType(data, str)
    interpolated = _CONFIG.InterpolateValue(data)
    # TODO(hanuszczak): This assertion should not be necessary but since the
    # whole configuration system is one gigantic spaghetti, we can never be sure
    # what is being returned.
    precondition.AssertType(data, str)
    return cast(str, interpolated)


class Flags(ConfigFilter):
  """Get the parameter from the flags."""

  name = "flags"

  def Filter(self, data: str):
    precondition.AssertType(data, str)
    try:
      logging.debug("Overriding config option with flags.FLAGS.%s", data)
      attribute = getattr(flags.FLAGS, data)
      # TODO(hanuszczak): Filters should always return strings and this juggling
      # should not be needed. This is just a quick hack to fix prod.
      if isinstance(attribute, bytes):
        attribute = attribute.decode("utf-8")
      elif not isinstance(attribute, str):
        attribute = str(attribute)
      # TODO(hanuszczak): See TODO comment in the `Expand` filter.
      precondition.AssertType(attribute, str)
      return cast(str, attribute)
    except AttributeError as e:
      raise FilterError(e)


class Resource(ConfigFilter):
  """Locates a GRR resource that is shipped with the GRR package.

  The format of the directive is "path/to/resource@package_name". If
  package_name is not provided we use grr-resource-core by default.
  """

  name = "resource"
  default_package = "grr-response-core"

  def Filter(self, data: str) -> str:
    """Use pkg_resources to find the path to the required resource."""
    if "@" in data:
      file_path, package_name = data.split("@")
    else:
      file_path, package_name = data, Resource.default_package

    resource_path = package.ResourcePath(package_name, file_path)
    if resource_path is not None:
      return resource_path

    # pylint: disable=unreachable
    raise FilterError("Unable to find resource %s while interpolating: " % data)
    # pylint: enable=unreachable


class ModulePath(ConfigFilter):
  """Locate the path to the specified module.

  Note: A module is either a python file (with a .py extension) or a directory
  with a __init__.py inside it. It is not the same as a resource (See Resource
  above) since a module will be installed somewhere you can import it from.

  Caveat: This will raise if the module is not a physically present on disk
  (e.g. pyinstaller bundle).
  """

  name = "module_path"

  def Filter(self, data: str) -> str:
    try:
      return package.ModulePath(data)
    except ImportError as e:
      message = (
          "Config parameter module_path expansion %r can not be imported."
          % data
      )

      # This exception will typically be caught by the expansion engine and
      # be silently swallowed.
      traceback.print_exc()
      logging.error(message)
      raise FilterError(message) from e


class StringInterpolator(lexer.Lexer):
  r"""Implements a lexer for the string interpolation language.

  Config files may specify nested interpolation codes:

  - The following form specifies an interpolation command:
      %(arg string|filter)

    Where arg string is an arbitrary string and filter is the name of a filter
    function which will receive the arg string. If filter is omitted, the arg
    string is interpreted as a section.parameter reference and expanded from
    within the config system.

  - Interpolation commands may be nested. In this case, the interpolation
    proceeds from innermost to outermost:

    e.g. %(arg1 %(arg2|filter2)|filter1)

      1. First arg2 is passed through filter2.
      2. The result of that is appended to arg1.
      3. The combined string is then filtered using filter1.

  - The following characters must be escaped by preceding them with a single \:
     - ()|
  """

  tokens = [
      # When in literal mode, only allow to escape }
      lexer.Token("Literal", r"\\[^{}]", "AppendArg", None),
      # Allow escaping of special characters
      lexer.Token(None, r"\\(.)", "Escape", None),
      # Literal sequence is %{....}. Literal states can not be nested further,
      # i.e. we include anything until the next }. It is still possible to
      # escape } if this character needs to be inserted literally.
      lexer.Token("Literal", r"\}", "EndLiteralExpression,PopState", None),
      lexer.Token("Literal", r"[^}\\]+", "AppendArg", None),
      lexer.Token(None, r"\%\{", "StartExpression,PushState", "Literal"),
      # Expansion sequence is %(....)
      lexer.Token(None, r"\%\(", "StartExpression", None),
      lexer.Token(None, r"\|([a-zA-Z_-]+)\)", "Filter", None),
      lexer.Token(None, r"\)", "ExpandArg", None),
      # Glob up as much data as possible to increase efficiency here.
      lexer.Token(None, r"[^()%{}|\\]+", "AppendArg", None),
      lexer.Token(None, r".", "AppendArg", None),
  ]

  STRING_ESCAPES = {
      "\\\\": "\\",
      "\\(": "(",
      "\\)": ")",
      "\\{": "{",
      "\\}": "}",
      "\\%": "%",
  }

  def __init__(
      self,
      data,
      config,
      default_section="",
      parameter=None,
      context=None,
  ):
    self.stack = [""]
    self.default_section = default_section
    self.parameter = parameter
    self.config = config
    self.context = context
    super().__init__(data)

  def Escape(self, string="", **_):
    """Support standard string escaping."""
    # Translate special escapes:
    self.stack[-1] += self.STRING_ESCAPES.get(string, string)

  def Error(self, message=None, weight=1):
    """Parse errors are fatal."""
    raise ConfigFormatError("While parsing %s: %s" % (self.parameter, message))

  def StartExpression(self, **_):
    """Start processing a new expression."""
    # Extend the stack for the new expression.
    self.stack.append("")

  def EndLiteralExpression(self, **_):
    if len(self.stack) <= 1:
      raise lexer.ParseError(
          "Unbalanced literal sequence: Can not expand '%s'"
          % self.processed_buffer
      )

    arg = self.stack.pop(-1)
    self.stack[-1] += arg

  def Filter(self, match=None, **_):
    """Filter the current expression."""
    arg = self.stack.pop(-1)

    # Filters can be specified as a comma separated list.
    for filter_name in match.group(1).split(","):
      filter_object = ConfigFilter.classes_by_name.get(filter_name)
      if filter_object is None:
        raise FilterError("Unknown filter function %r" % filter_name)

      if not filter_object.sensitive_arg:
        logging.debug("Applying filter %s for %s.", filter_name, arg)
      arg = filter_object().Filter(arg)
      precondition.AssertType(arg, str)

    self.stack[-1] += arg

  def ExpandArg(self, **_):
    """Expand the args as a section.parameter from the config."""
    # This function is called when we see close ) and the stack depth has to
    # exactly match the number of (.
    if len(self.stack) <= 1:
      raise lexer.ParseError(
          "Unbalanced parenthesis: Can not expand '%s'" % self.processed_buffer
      )

    # This is the full parameter name: e.g. Logging.path
    parameter_name = self.stack.pop(-1)
    if "." not in parameter_name:
      parameter_name = "%s.%s" % (self.default_section, parameter_name)

    final_value = self.config.Get(parameter_name, context=self.context)
    if final_value is None:
      final_value = ""

    type_info_obj = (
        self.config.FindTypeInfo(parameter_name) or type_info.String()
    )

    # Encode the interpolated string according to its type.
    self.stack[-1] += type_info_obj.ToString(final_value)

  def AppendArg(self, string="", **_):
    self.stack[-1] += string

  def Parse(self):
    self.Close()
    if len(self.stack) != 1:
      raise lexer.ParseError("Nested expression not balanced.")

    return self.stack[0]


class GrrConfigManager(object):
  """Manage configuration system in GRR."""

  def __init__(self):
    """Initialize the configuration manager."""
    # The context is used to provide different configuration directives in
    # different situations. The context can contain any string describing a
    # different aspect of the running instance.
    self.context = []
    self.raw_data = dict()
    self.files = []
    self.secondary_config_parsers = []
    self.writeback = None
    self.writeback_data = dict()
    self.global_override = dict()
    self.context_descriptions = {}
    self.constants = set()
    self.valid_contexts = set()

    # This is the type info set describing all configuration
    # parameters.
    self.type_infos = type_info.TypeDescriptorSet()

    # We store the defaults here.
    self.defaults = {}

    # A cache of validated and interpolated results.
    self.FlushCache()

    self.initialized = False
    self.DeclareBuiltIns()

  def DeclareBuiltIns(self):
    """Declare built in options internal to the config system."""
    self.DEFINE_list(
        "Config.includes",
        [],
        "List of additional config files to include. Files are "
        "processed recursively depth-first, later values "
        "override earlier ones.",
    )

  def __str__(self):
    # List all the files we read from.
    message = ""
    for filename in self.files:
      message += ' file="%s" ' % filename

    return "<%s %s>" % (self.__class__.__name__, message)

  def FlushCache(self):
    self.cache = {}

  def MakeNewConfig(self):
    """Creates a new configuration option based on this one.

    Note that it is not normally possible to just instantiate the
    config object because it will have an empty set of type
    descriptors (i.e. no config options will be defined). Config
    options are normally defined at import time, and then they get
    added to the _CONFIG global in this module.

    To obtain a new configuration object, inheriting the regular
    config options, this method must be called from the global _CONFIG
    object, to make a copy.

    Returns:
      A new empty config object. which has the same parameter definitions as
      this one.
    """
    result = self.__class__()

    # We do not need to copy here since these never change.
    result.type_infos = self.type_infos
    result.defaults = self.defaults
    result.context = self.context
    result.valid_contexts = self.valid_contexts

    return result

  def CopyConfig(self):
    """Make a complete new copy of the current config.

    This includes all options as they currently are. If you want a base config
    with defaults use MakeNewConfig.

    Returns:
      A new config object with the same data as self.
    """
    newconf = self.MakeNewConfig()
    newconf.raw_data = copy.deepcopy(self.raw_data)
    newconf.files = copy.deepcopy(self.files)
    newconf.secondary_config_parsers = [
        p.Copy() for p in self.secondary_config_parsers
    ]
    newconf.writeback = copy.deepcopy(self.writeback)
    newconf.writeback_data = copy.deepcopy(self.writeback_data)
    newconf.global_override = copy.deepcopy(self.global_override)
    newconf.context_descriptions = copy.deepcopy(self.context_descriptions)
    newconf.constants = copy.deepcopy(self.constants)
    newconf.initialized = copy.deepcopy(self.initialized)
    return newconf

  def SetWriteBack(self, filename, rename_invalid_writeback=True):
    """Sets the config file which will receive any modifications.

    The main config file can be made writable, but directing all Set()
    operations into a secondary location. This secondary location will
    receive any updates and will override the options for this file.

    Args:
      filename: A filename which will receive updates. The file is parsed first
        and merged into the raw data from this object.
      rename_invalid_writeback: Whether to rename the writeback file if it
        cannot be parsed.
    """
    try:
      self.writeback = self.LoadSecondaryConfig(filename)
      self.MergeData(self.writeback.ReadData(), self.writeback_data)
    except config_parser.ReadDataError as e:
      # This means that we probably aren't installed correctly.
      logging.error("Unable to read writeback file: %s", e)
      return
    except Exception as we:  # pylint: disable=broad-except
      # Could be yaml parse error, could be some malformed parameter. Move the
      # writeback file so that we start in a clean state next run
      if rename_invalid_writeback and os.path.exists(filename):
        try:
          b = filename + ".bak"
          os.rename(filename, b)
          logging.warning("Broken writeback (%s) renamed to: %s", we, b)
        except Exception as e:  # pylint: disable=broad-except
          logging.error("Unable to rename broken writeback: %s", e)
      raise we
    logging.debug("Configuration writeback is set to %s", filename)

  def Validate(self, sections=None, parameters=None):
    """Validate sections or individual parameters.

    The GRR configuration file contains several sections, used by different
    components. Many of these components don't care about other sections. This
    method allows a component to declare in advance what sections and parameters
    it cares about, and have these validated.

    Args:
      sections: A list of sections to validate. All parameters within the
        section are validated.
      parameters: A list of specific parameters (in the format section.name) to
        validate.

    Returns:
      dict of {parameter: Exception}, where parameter is a section.name string.
    """
    if isinstance(sections, str):
      sections = [sections]

    if sections is None:
      sections = []

    if parameters is None:
      parameters = []

    validation_errors = {}
    for section in sections:
      for descriptor in self.type_infos:
        if descriptor.name.startswith(section + "."):
          try:
            self.Get(descriptor.name)
          except (Error, ValueError) as e:
            validation_errors[descriptor.name] = e

    for parameter in parameters:
      for descriptor in self.type_infos:
        if parameter == descriptor.name:
          try:
            self.Get(descriptor.name)
          except (Error, ValueError) as e:
            validation_errors[descriptor.name] = e

    return validation_errors

  def AddContext(self, context_string, description=None):
    """Adds a context string to the global configuration.

    The context conveys information about the caller of the config system and
    allows the configuration to have specialized results for different callers.

    Note that the configuration file may specify conflicting options for
    different contexts. In this case, later specified contexts (i.e. the later
    AddContext() calls) will trump the earlier specified contexts. This allows
    specialized contexts to be specified on the command line which override
    normal operating options.

    Args:
      context_string: A string which describes the global program.
      description: A description as to when this context applies.

    Raises:
      InvalidContextError: An undefined context was specified.
    """
    if context_string not in self.context:
      if context_string not in self.valid_contexts:
        raise InvalidContextError(
            "Invalid context specified: %s" % context_string
        )

      self.context.append(context_string)
      self.context_descriptions[context_string] = description

    self.FlushCache()

  def ContextApplied(self, context_string):
    """Return true if the context is applied."""
    return context_string in self.context

  def RemoveContext(self, context_string):
    if context_string in self.context:
      self.context.remove(context_string)
      self.context_descriptions.pop(context_string)

    self.FlushCache()

  def SetRaw(self, name, value):
    """Set the raw string without verification or escaping."""
    if self.writeback is None:
      logging.warning("Attempting to modify a read only config object.")
    if name in self.constants:
      raise ConstModificationError(
          "Attempting to modify constant value %s" % name
      )

    self.writeback_data[name] = value
    self.FlushCache()

  def Set(self, name, value):
    """Update the configuration option with a new value.

    Note that this forces the value to be set for all contexts. The value is
    written to the writeback location if Save() is later called.

    Args:
      name: The name of the parameter to set.
      value: The value to set it to. The value will be validated against the
        option's type descriptor.

    Raises:
      ConstModificationError: When attempting to change a constant option.
    """
    # If the configuration system has a write back location we use it,
    # otherwise we use the primary configuration object.
    if self.writeback is None:
      logging.warning(
          "Attempting to modify a read only config object for %s.", name
      )
    if name in self.constants:
      raise ConstModificationError(
          "Attempting to modify constant value %s" % name
      )

    writeback_data = self.writeback_data

    # Check if the new value conforms with the type_info.
    if value is not None:
      if isinstance(value, str):
        value = self.EscapeString(value)

      if isinstance(value, bytes):
        raise ValueError(
            "Setting config option %s to bytes is not allowed" % name
        )

    writeback_data[name] = value
    self.FlushCache()

  def EscapeString(self, string):
    """Escape special characters when encoding to a string."""
    return re.sub(r"([\\%){}])", r"\\\1", string)

  def Write(self):
    """Write out the updated configuration to the fd."""
    if self.writeback:
      self.writeback.SaveData(self.writeback_data)
    else:
      raise RuntimeError(
          "Attempting to write a configuration without a writeback location."
      )

  def Persist(self, config_option):
    """Stores <config_option> in the writeback."""
    if not self.writeback:
      raise RuntimeError(
          "Attempting to write a configuration without a writeback location."
      )

    writeback_raw_value = dict(self.writeback.ReadData()).get(config_option)
    raw_value = None

    for parser in [self.parser] + self.secondary_config_parsers:
      if parser == self.writeback:
        continue

      config_raw_data = dict(parser.ReadData())
      raw_value = config_raw_data.get(config_option)
      if raw_value is None:
        continue
      break

    if writeback_raw_value == raw_value:
      return

    if raw_value is None:
      return

    self.SetRaw(config_option, raw_value)
    self.Write()

  def AddOption(self, descriptor, constant=False):
    """Registers an option with the configuration system.

    Args:
      descriptor: A TypeInfoObject instance describing the option.
      constant: If this is set, the option is treated as a constant - it can be
        read at any time (before parsing the configuration) and it's an error to
        try to override it in a config file.

    Raises:
      RuntimeError: The descriptor's name must contain a . to denote the section
         name, otherwise we raise.
      AlreadyInitializedError: If the config has already been read it's too late
         to define new options.
    """
    if self.initialized:
      raise AlreadyInitializedError(
          "Config was already initialized when defining %s" % descriptor.name
      )

    if descriptor.name in self.type_infos:
      logging.warning("Config Option %s multiply defined!", descriptor.name)

    self.type_infos.Append(descriptor)
    if constant:
      self.constants.add(descriptor.name)

    # Register this option's default value.
    self.defaults[descriptor.name] = descriptor.GetDefault()
    self.FlushCache()

  def DefineContext(self, context_name):
    self.valid_contexts.add(context_name)
    return context_name

  def FormatHelp(self):
    result = "Context: %s\n\n" % ",".join(self.context)
    for descriptor in sorted(self.type_infos, key=lambda x: x.name):
      result += descriptor.Help() + "\n"
      try:
        result += "   Current Value: %s\n" % self.Get(descriptor.name)
      except Exception as e:  # pylint:disable=broad-except
        result += "   Current Value: %s (Error: %s)\n" % (
            self.GetRaw(descriptor.name),
            e,
        )
    return result

  def PrintHelp(self):
    print(self.FormatHelp())

  def MergeData(self, merge_data: dict[Any, Any], raw_data=None):
    """Merges data read from a config file into the current config."""
    self.FlushCache()
    if raw_data is None:
      raw_data = self.raw_data

    for k, v in merge_data.items():
      # A context clause.
      if isinstance(v, dict) and k not in self.type_infos:
        if k not in self.valid_contexts:
          raise InvalidContextError("Invalid context specified: %s" % k)
        context_data = raw_data.setdefault(k, dict())
        self.MergeData(v, context_data)

      else:
        # Find the descriptor for this field.
        descriptor = self.type_infos.get(k)
        if descriptor is None:
          msg = (
              "Missing config definition for %s. This option is likely "
              "deprecated or renamed. Check the release notes." % k
          )
          if flags.FLAGS.disallow_missing_config_definitions:
            raise MissingConfigDefinitionError(msg)

        if isinstance(v, str):
          v = v.strip()

        # If we are already initialized and someone tries to modify a constant
        # value (e.g. via Set()), break loudly.
        if self.initialized and k in self.constants:
          raise ConstModificationError(
              "Attempting to modify constant value %s" % k
          )

        raw_data[k] = v

  def LoadSecondaryConfig(
      self, filename=None, parser=None, process_includes=True
  ) -> config_parser.GRRConfigParser:
    """Loads an additional configuration file.

    The configuration system has the concept of a single Primary configuration
    file, and multiple secondary files. The primary configuration file is the
    main file that is used by the program. Any writebacks will only be made to
    the primary configuration file. Secondary files contain additional
    configuration data which will be merged into the configuration system.

    This method adds an additional configuration file.

    Args:
      filename: The configuration file that will be loaded. For example
           file:///etc/grr.conf or reg://HKEY_LOCAL_MACHINE/Software/GRR.
      parser: An optional parser can be given. In this case, the parser's data
        will be loaded directly.
      process_includes: If false, do not process any files listed in
        Config.includes configuration option.

    Returns:
      The parser used to parse this configuration source.

    Raises:
      ValueError: if both filename and parser arguments are None.
      ConfigFileNotFound: If a specified included file was not found.
    """
    if filename:
      # Maintain a stack of config file locations in loaded order.
      self.files.append(filename)

      parser = config_parser.GetParserFromPath(filename)
      logging.debug("Loading configuration from %s", filename)
      self.secondary_config_parsers.append(parser)
    elif parser is None:
      raise ValueError("Must provide either a filename or a parser.")

    clone = self.MakeNewConfig()
    clone.MergeData(parser.ReadData())
    clone.initialized = True

    if process_includes:
      for file_to_load in clone["Config.includes"]:
        # We can not include a relative file from a config which does not have
        # path.
        if not os.path.isabs(file_to_load):
          if not filename:
            raise ConfigFileNotFound(
                "While loading %s: Unable to include a relative path (%s) "
                "from a config without a filename" % (filename, file_to_load)
            )

          # If the included path is relative, we take it as relative to the
          # current path of the config.
          file_to_load = os.path.join(os.path.dirname(filename), file_to_load)

        clone_parser = clone.LoadSecondaryConfig(file_to_load)
        # If an include file is specified but it was not found, raise an error.
        try:
          clone_parser.ReadData()
        except config_parser.ReadDataError as e:
          raise ConfigFileNotFound(
              "Unable to load include file %s" % file_to_load
          ) from e

    self.MergeData(clone.raw_data)
    self.files.extend(clone.files)

    return parser

  # TODO(hanuszczak): Magic method with a lot of mutually exclusive switches. It
  # should be split into multiple methods instead.
  def Initialize(
      self,
      filename: Optional[str] = None,
      data: Optional[str] = None,
      fd: Optional[BinaryIO] = None,
      reset: bool = True,
      must_exist: bool = False,
      process_includes: bool = True,
      parser: type[
          config_parser.GRRConfigParser
      ] = config_parser.IniConfigFileParser,
  ):
    """Initializes the config manager.

    This method is used to add more config options to the manager. The config
    can be given as one of the parameters as described in the Args section.

    Args:
      filename: The name of the configuration file to use.
      data: The configuration given directly as a long string of data.
      fd: A file descriptor of a configuration file.
      reset: If true, the previous configuration will be erased.
      must_exist: If true the data source must exist and be a valid
        configuration file, or we raise an exception.
      process_includes: If false, do not process any files listed in
        Config.includes configuration option.
      parser: The parser class to use (i.e. the format of the file). If not
        specified guess from the filename.

    Raises:
      RuntimeError: No configuration was passed in any of the parameters.

      ConfigFormatError: Raised when the configuration file is invalid or does
        not exist..
    """
    self.FlushCache()
    if reset:
      # Clear previous configuration.
      self.raw_data = dict()
      self.writeback_data = dict()
      self.writeback = None
      self.initialized = False

    if fd is not None:
      if issubclass(parser, config_parser.GRRConfigFileParser):
        self.parser = self.LoadSecondaryConfig(
            parser=config_parser.FileParserDataWrapper(fd.read(), parser("")),
            process_includes=process_includes,
        )
      else:
        raise TypeError("Trying to read from FD with a non-file parser.")

    elif filename is not None:
      self.parser = self.LoadSecondaryConfig(
          filename, process_includes=process_includes
      )
      try:
        self.parser.ReadData()
      except config_parser.ReadDataError as e:
        if must_exist:
          raise ConfigFormatError(
              "Unable to parse config file %s" % filename
          ) from e

    elif data is not None:
      if issubclass(parser, config_parser.GRRConfigFileParser):
        self.parser = self.LoadSecondaryConfig(
            parser=config_parser.FileParserDataWrapper(
                data.encode("utf-8"), parser("")
            ),
            process_includes=process_includes,
        )
      else:
        raise TypeError("Trying to parse bytes with a non-file parser.")

    elif must_exist:
      raise RuntimeError("Registry path not provided.")

    self.initialized = True

  def __getitem__(self, name):
    """Retrieve a configuration value after suitable interpolations."""
    if name not in self.type_infos:
      raise UnknownOption("Config parameter %s not known." % name)

    return self.Get(name)

  def GetRaw(self, name, context=None, default=utils.NotAValue):
    """Get the raw value without interpolations."""
    if context is None:
      context = self.context

    # Getting a raw value is pretty cheap so we won't bother with the cache
    # here.
    _, value = self._GetValue(name, context, default=default)
    return value

  def Get(self, name, default=utils.NotAValue, context=None):
    """Get the value contained  by the named parameter.

    This method applies interpolation/escaping of the named parameter and
    retrieves the interpolated value.

    Args:
      name: The name of the parameter to retrieve. This should be in the format
        of "Section.name"
      default: If retrieving the value results in an error, return this default.
      context: A list of context strings to resolve the configuration. This is a
        set of roles the caller is current executing with. For example (client,
        windows). If not specified we take the context from the current thread's
        TLS stack.

    Returns:
      The value of the parameter.
    Raises:
      ConfigFormatError: if verify=True and the config doesn't validate.
      RuntimeError: if a value is retrieved before the config is initialized.
      ValueError: if a bad context is passed.
    """
    if not self.initialized:
      if name not in self.constants:
        raise RuntimeError(
            "Error while retrieving %s: "
            "Configuration hasn't been initialized yet." % name
        )
    if context:
      # Make sure it's not just a string and is iterable.
      if isinstance(context, str) or not isinstance(context, abc.Iterable):
        raise ValueError("context should be a list, got %r" % context)

    calc_context = context

    # Only use the cache if possible.
    cache_key = (name, tuple(context or ()))
    if default is utils.NotAValue and cache_key in self.cache:
      return self.cache[cache_key]

    # Use a default global context if context is not provided.
    if context is None:
      calc_context = self.context

    type_info_obj = self.FindTypeInfo(name)
    _, return_value = self._GetValue(
        name, context=calc_context, default=default
    )

    # If we returned the specified default, we just return it here.
    if return_value is default:
      return default

    try:
      return_value = self.InterpolateValue(
          return_value,
          default_section=name.split(".")[0],
          type_info_obj=type_info_obj,
          context=calc_context,
      )
    except (lexer.ParseError, ValueError) as e:
      # We failed to parse the value, but a default was specified, so we just
      # return that.
      if default is not utils.NotAValue:
        return default

      raise ConfigFormatError("While parsing %s: %s" % (name, e))

    try:
      new_value = type_info_obj.Validate(return_value)
      if new_value is not None:
        # Update the stored value with the valid data.
        return_value = new_value
    except ValueError:
      if default is not utils.NotAValue:
        return default

      raise

    # Cache the value for next time.
    if default is utils.NotAValue:
      self.cache[cache_key] = return_value

    return return_value

  def _ResolveContext(self, context, name, raw_data, path=None):
    """Returns the config options active under this context."""
    if path is None:
      path = []

    for element in context:
      if element not in self.valid_contexts:
        raise InvalidContextError("Invalid context specified: %s" % element)

      if element in raw_data:
        context_raw_data = raw_data[element]

        # TODO(hanuszczak): Investigate why pytype complains here (probably for
        # valid reasons, because this code does not looks like something well-
        # typed).
        value = context_raw_data.get(name)  # pytype: disable=attribute-error
        if value is not None:
          if isinstance(value, str):
            value = value.strip()

          yield context_raw_data, value, path + [element]

        # Recurse into the new context configuration.
        for context_raw_data, value, new_path in self._ResolveContext(
            context, name, context_raw_data, path=path + [element]
        ):
          yield context_raw_data, value, new_path

  def _GetValue(self, name, context, default=utils.NotAValue):
    """Search for the value based on the context."""
    container = self.defaults

    # The caller provided a default value.
    if default is not utils.NotAValue:
      value = default

    # Take the default from the definition.
    elif name in self.defaults:
      value = self.defaults[name]

    else:
      raise UnknownOption("Option %s not defined." % name)

    # We resolve the required key with the default raw data, and then iterate
    # over all elements in the context to see if there are overriding context
    # configurations.
    new_value = self.raw_data.get(name)
    if new_value is not None:
      value = new_value
      container = self.raw_data

    # Now check for any contexts. We enumerate all the possible resolutions and
    # sort by their path length. The last one will be the match with the deepest
    # path (i.e .the most specific match).
    matches = list(self._ResolveContext(context, name, self.raw_data))

    if matches:
      # Sort by the length of the path - longest match path will be at the end.
      matches.sort(key=lambda x: len(x[2]))
      value = matches[-1][1]
      container = matches[-1][0]

      if (
          len(matches) >= 2
          and len(matches[-1][2]) == len(matches[-2][2])
          and matches[-1][2] != matches[-2][2]
          and matches[-1][1] != matches[-2][1]
      ):
        # This warning specifies that there is an ambiguous match, the config
        # attempts to find the most specific value e.g. if you have a value
        # for X.y in context A,B,C, and a value for X.y in D,B it should choose
        # the one in A,B,C. This warning is for if you have a value in context
        # A,B and one in A,C. The config doesn't know which one to pick so picks
        # one and displays this warning.
        logging.warning(
            "Ambiguous configuration for key %s: "
            "Contexts of equal length: %s (%s) and %s (%s)",
            name,
            matches[-1][2],
            matches[-1][1],
            matches[-2][2],
            matches[-2][1],
        )

    # If there is a writeback location this overrides any previous
    # values.
    if self.writeback_data:
      new_value = self.writeback_data.get(name)
      if new_value is not None:
        value = new_value
        container = self.writeback_data

    # Allow the global override to force an option value.
    if name in self.global_override:
      return self.global_override, self.global_override[name]

    return container, value

  def FindTypeInfo(self, name):
    """Search for a type_info instance which describes this key."""
    result = self.type_infos.get(name)
    if result is None:
      # Not found, assume string.
      result = type_info.String(name=name, default="")

    return result

  def InterpolateValue(
      self,
      value,
      type_info_obj=type_info.String(),
      default_section=None,
      context=None,
  ):
    """Interpolate the value and parse it with the appropriate type."""
    # It is only possible to interpolate strings...
    if isinstance(value, str):
      try:
        value = StringInterpolator(
            value,
            self,
            default_section=default_section,
            parameter=type_info_obj.name,
            context=context,
        ).Parse()
      except InterpolationError as e:
        # TODO(hanuszczak): This is a quick hack to not refactor too much while
        # working on Python 3 compatibility. But this is bad and exceptions
        # should not be used like this.
        message = "{cause}: {detail}".format(cause=e, detail=value)
        raise type(e)(message)

      # Parse the data from the string.
      value = type_info_obj.FromString(value)

    # ... and lists of strings.
    if isinstance(value, list):
      value = [
          self.InterpolateValue(
              v, default_section=default_section, context=context
          )
          for v in value
      ]

    return value

  def GetSections(self):
    result = set()
    for descriptor in self.type_infos:
      result.add(descriptor.section)

    return result

  def MatchBuildContext(
      self,
      target_os,
      target_arch,
      target_package,
      context=None,
  ):
    """Return true if target_platforms matches the supplied parameters.

    Used by buildanddeploy to determine what clients need to be built.

    Args:
      target_os: which os we are building for in this run (linux, windows,
        darwin)
      target_arch: which arch we are building for in this run (i386, amd64)
      target_package: which package type we are building (exe, dmg, deb, rpm)
      context: config_lib context

    Returns:
      bool: True if target_platforms spec matches parameters.
    """
    for spec in self.Get("ClientBuilder.target_platforms", context=context):
      spec_os, arch, package_name = spec.split("_")
      if (
          spec_os == target_os
          and arch == target_arch
          and package_name == target_package
      ):
        return True
    return False

  # pylint: disable=g-bad-name,redefined-builtin
  def DEFINE_bool(self, name, default, help, constant=False):
    """A helper for defining boolean options."""
    self.AddOption(
        type_info.Bool(name=name, default=default, description=help),
        constant=constant,
    )

  def DEFINE_float(self, name, default, help, constant=False):
    """A helper for defining float options."""
    self.AddOption(
        type_info.Float(name=name, default=default, description=help),
        constant=constant,
    )

  def DEFINE_integer(self, name, default, help, constant=False):
    """A helper for defining integer options."""
    self.AddOption(
        type_info.Integer(name=name, default=default, description=help),
        constant=constant,
    )

  def DEFINE_string(self, name, default, help, constant=False):
    """A helper for defining string options."""
    self.AddOption(
        type_info.String(name=name, default=default or "", description=help),
        constant=constant,
    )

  def DEFINE_choice(self, name, default, choices, help, constant=False):
    """A helper for defining choice string options."""
    self.AddOption(
        type_info.Choice(
            name=name, default=default, choices=choices, description=help
        ),
        constant=constant,
    )

  def DEFINE_multichoice(self, name, default, choices, help, constant=False):
    """Choose multiple options from a list."""
    self.AddOption(
        type_info.MultiChoice(
            name=name, default=default, choices=choices, description=help
        ),
        constant=constant,
    )

  def DEFINE_integer_list(self, name, default, help, constant=False):
    """A helper for defining lists of integer options."""
    self.AddOption(
        type_info.List(
            name=name,
            default=default,
            description=help,
            validator=type_info.Integer(),
        ),
        constant=constant,
    )

  def DEFINE_list(self, name, default, help, constant=False):
    """A helper for defining lists of strings options."""
    self.AddOption(
        type_info.List(
            name=name,
            default=default,
            description=help,
            validator=type_info.String(),
        ),
        constant=constant,
    )

  def DEFINE_constant_string(self, name, default, help):
    """A helper for defining constant strings."""
    self.AddOption(
        type_info.String(name=name, default=default or "", description=help),
        constant=True,
    )

  def DEFINE_semantic_value(self, semantic_type, name, default=None, help=""):
    if issubclass(semantic_type, rdf_structs.RDFStruct):
      raise ValueError(
          "DEFINE_semantic_value should be used for types based on primitives."
      )

    self.AddOption(
        type_info.RDFValueType(
            rdfclass=semantic_type, name=name, default=default, description=help
        )
    )

  def DEFINE_semantic_enum(
      self,
      enum_container: rdf_structs.EnumContainer,
      name: str,
      default: Optional[rdf_structs.EnumNamedValue] = None,
      help: str = "",
  ) -> None:
    if not isinstance(enum_container, rdf_structs.EnumContainer):
      raise ValueError("enum_container must be an EnumContainer.")

    self.AddOption(
        type_info.RDFEnumType(
            enum_container=enum_container,
            name=name,
            default=default,
            description=help,
        )
    )

  def DEFINE_semantic_struct(self, semantic_type, name, default=None, help=""):
    if not issubclass(semantic_type, rdf_structs.RDFStruct):
      raise ValueError(
          "DEFINE_semantic_struct should be used for types based on structs."
      )

    self.AddOption(
        type_info.RDFStructDictType(
            rdfclass=semantic_type, name=name, default=default, description=help
        )
    )

  def DEFINE_context(self, name):
    return self.DefineContext(name)


# pylint: enable=g-bad-name

# Global config object. This object is not supposed to be used directly,
# since when used from config_lib, it's not going to have all the GRR
# config options registered.
#
# grr.config.CONFIG should be used instead - using grr.config.CONFIG
# guarantees that all configuration options and filters are properly
# imported.
_CONFIG = GrrConfigManager()


# pylint: disable=g-bad-name,redefined-builtin
def DEFINE_bool(name, default, help):
  """A helper for defining boolean options."""
  _CONFIG.DEFINE_bool(name, default, help)


def DEFINE_float(name, default, help):
  """A helper for defining float options."""
  _CONFIG.DEFINE_float(name, default, help)


def DEFINE_integer(name, default, help):
  """A helper for defining integer options."""
  _CONFIG.DEFINE_integer(name, default, help)


def DEFINE_boolean(name, default, help):
  """A helper for defining boolean options."""
  _CONFIG.DEFINE_bool(name, default, help)


def DEFINE_string(name, default, help):
  """A helper for defining string options."""
  _CONFIG.DEFINE_string(name, default, help)


def DEFINE_choice(name, default, choices, help):
  """A helper for defining choice string options."""
  _CONFIG.DEFINE_choice(name, default, choices, help)


def DEFINE_multichoice(name, default, choices, help):
  """Choose multiple options from a list."""
  _CONFIG.DEFINE_multichoice(name, default, choices, help)


def DEFINE_integer_list(name, default, help):
  """A helper for defining lists of integer options."""
  _CONFIG.DEFINE_integer_list(name, default, help)


def DEFINE_list(name, default, help):
  """A helper for defining lists of strings options."""
  _CONFIG.DEFINE_list(name, default, help)


def DEFINE_semantic_value(semantic_type, name, default=None, help=""):
  _CONFIG.DEFINE_semantic_value(semantic_type, name, default=default, help=help)


def DEFINE_semantic_enum(
    semantic_enum: rdf_structs.EnumContainer,
    name: str,
    default: Optional[rdf_structs.EnumNamedValue] = None,
    help: str = "",
) -> None:
  _CONFIG.DEFINE_semantic_enum(semantic_enum, name, default=default, help=help)


def DEFINE_semantic_struct(semantic_type, name, default=None, help=""):
  _CONFIG.DEFINE_semantic_struct(
      semantic_type, name, default=default, help=help
  )


def DEFINE_option(type_descriptor):
  _CONFIG.AddOption(type_descriptor)


def DEFINE_constant_string(name, default, help):
  """A helper for defining constant strings."""
  _CONFIG.DEFINE_constant_string(name, default, help)


def DEFINE_context(name):
  return _CONFIG.DefineContext(name)


# pylint: enable=g-bad-name


def LoadConfig(
    config_obj,
    config_file=None,
    config_fd=None,
    secondary_configs=None,
    contexts=None,
    reset=False,
    parser=config_parser.IniConfigFileParser,
):
  """Initialize a ConfigManager with the specified options.

  Args:
    config_obj: The ConfigManager object to use and update. If None, one will be
      created.
    config_file: Filename to read the config from.
    config_fd: A file-like object to read config data from.
    secondary_configs: A list of secondary config URLs to load.
    contexts: Add these contexts to the config object.
    reset: Completely wipe previous config before doing the load.
    parser: Specify which parser to use.

  Returns:
    The resulting config object. The one passed in, unless None was specified.
  """
  if config_obj is None or reset:
    # Create a new config object.
    config_obj = _CONFIG.MakeNewConfig()

  # Initialize the config with a filename or file like object.
  if config_file is not None:
    config_obj.Initialize(filename=config_file, must_exist=True, parser=parser)
  elif config_fd is not None:
    config_obj.Initialize(fd=config_fd, parser=parser)

  # Load all secondary files.
  if secondary_configs:
    for config_file in secondary_configs:
      config_obj.LoadSecondaryConfig(config_file)

  if contexts:
    for context in contexts:
      config_obj.AddContext(context)

  return config_obj


def ParseConfigCommandLine(rename_invalid_writeback=True):
  """Parse all the command line options which control the config system."""
  # The user may specify the primary config file on the command line.
  if flags.FLAGS.config:
    _CONFIG.Initialize(filename=flags.FLAGS.config, must_exist=True)
  else:
    raise RuntimeError("A config file is not specified.")

  # Allow secondary configuration files to be specified.
  if flags.FLAGS.secondary_configs:
    for config_file in flags.FLAGS.secondary_configs:
      _CONFIG.LoadSecondaryConfig(config_file)

  # Allow individual options to be specified as global overrides.
  for statement in flags.FLAGS.parameter:
    if "=" not in statement:
      raise RuntimeError("statement %s on command line not valid." % statement)

    name, value = statement.split("=", 1)
    _CONFIG.global_override[name] = value

  # Load additional contexts from the command line.
  for context in flags.FLAGS.context:
    if context:
      _CONFIG.AddContext(context)

  if _CONFIG["Config.writeback"]:
    _CONFIG.SetWriteBack(
        _CONFIG["Config.writeback"],
        rename_invalid_writeback=rename_invalid_writeback,
    )

  # Does the user want to dump help? We do this after the config system is
  # initialized so the user can examine what we think the value of all the
  # parameters are.
  if flags.FLAGS.config_help:
    print("Configuration overview.")

    _CONFIG.PrintHelp()
    sys.exit(0)
